import copy

import numpy as np
from sklearn.model_selection import GridSearchCV
from sklearn.neural_network import MLPClassifier

from mcalf.models.base import BASE_ATTRIBUTES, BASE_PARAMETERS, ModelBase
from mcalf.models.results import FitResult
from mcalf.profiles.voigt import double_voigt_nobg, voigt_nobg
from mcalf.utils.collections import OrderedParameterDict, Parameter
from mcalf.utils.misc import load_parameter, update_signature
from mcalf.utils.spec import generate_sigma
from mcalf.visualisation import plot_ibis8542

__all__ = ['IBIS8542Model']


class IBIS8542Model(ModelBase):
    """Class for working with IBIS 8542 Å calcium II spectral imaging observations.

    Parameters
    ----------
    ${PARAMETERS}

    Attributes
    ----------
    ${ATTRIBUTES}
    quiescent_wavelength : int, default=1
        The index within the fitted parameters of the absorption Voigt line core wavelength.
    active_wavelength : int, default=5
        The index within the fitted parameters of the emission Voigt line core wavelength.
    ${ATTRIBUTES_EXTRA}
    """

    default_kwargs = default_ibis8542model_kwargs = OrderedParameterDict([
        ('stationary_line_core', Parameter('stationary_line_core', 8542.099145376844)),
        ('absorption_guess', [-1000, Parameter('stationary_line_core'), 0.2, 0.1]),
        ('emission_guess', [1000, Parameter('stationary_line_core'), 0.2, 0.1]),
        ('absorption_min_bound', [-np.inf, Parameter('stationary_line_core') - 0.15, 1e-6, 1e-6]),
        ('emission_min_bound', [0, -np.inf, 1e-6, 1e-6]),
        ('absorption_max_bound', [0, Parameter('stationary_line_core') + 0.15, 1, 1]),
        ('emission_max_bound', [np.inf, np.inf, 1, 1]),
        ('absorption_x_scale', [1500, 0.2, 0.3, 0.5]),
        ('emission_x_scale', [1500, 0.2, 0.3, 0.5]),
        ('random_state', None),
    ])

    def __init__(self, **kwargs):

        # STAGE 1: Define dictionary of default attribute values
        defaults = copy.deepcopy(self.default_ibis8542model_kwargs)

        # STAGE 1.1: Initialise the parent class `ModelBase`
        class_keys = set(defaults.keys()) - {'stationary_line_core'}  # Keys that should not be passed to parent class
        base_kwargs = {k: kwargs[k] for k in kwargs.keys() if k not in class_keys}
        if 'stationary_line_core' not in kwargs.keys():  # Pass the child's default value if not specified
            base_kwargs['stationary_line_core'] = defaults['stationary_line_core'].eval()
        super().__init__(**base_kwargs)

        # STAGE 1.2: Load child default values for parent class attributes
        # stationary_line_core
        defaults.update_parameter('stationary_line_core', self.stationary_line_core)
        defaults = defaults.eval()  # Fix the 'stationary_line_core' value
        # TODO: Allow 'stationary_line_core' to affect other parameters dynamically after this point.
        # prefilter_response
        self._set_prefilter()  # Update the prefilter using stationary_line_core
        # sigma
        if self.sigma is None or (isinstance(self.sigma, bool) and self.sigma):
            self.sigma = [generate_sigma(i, self.constant_wavelengths, self.stationary_line_core) for i in [1, 2]]
        elif isinstance(self.sigma, bool) and not self.sigma:
            self.sigma = [np.ones(len(self.constant_wavelengths)) for i in [1, 2]]

        # STAGE 2: Update defaults with any values specified in a config file
        class_defaults = {k: self.config[k] for k in self.config.keys() if k in class_keys}
        for k in class_defaults.keys():
            if k in ['absorption_x_scale', 'emission_x_scale', 'random_state']:
                # These should not need the stationary line core
                class_defaults[k] = load_parameter(class_defaults[k])
            else:
                class_defaults[k] = load_parameter(class_defaults[k], wl=self.stationary_line_core)
            self.config.pop(k)  # Remove copied parameter
        defaults.update(class_defaults)  # Update the defaults with the config file

        # STAGE 3: Update defaults with the keyword arguments passed into the class initialisation
        class_kwargs = {k: kwargs[k] for k in class_keys if k in kwargs.keys()}
        defaults.update(class_kwargs)  # Update the defaults

        # STAGE 4: Set the object attributes (with some type enforcing)
        # values in the defaults dict
        self.absorption_guess = list(defaults['absorption_guess'])
        self.emission_guess = list(defaults['emission_guess'])
        self.absorption_min_bound = list(defaults['absorption_min_bound'])
        self.emission_min_bound = list(defaults['emission_min_bound'])
        self.absorption_max_bound = list(defaults['absorption_max_bound'])
        self.emission_max_bound = list(defaults['emission_max_bound'])
        self.absorption_x_scale = list(defaults['absorption_x_scale'])
        self.emission_x_scale = list(defaults['emission_x_scale'])
        # attributes whose default value cannot be changed during initialisation
        self.quiescent_wavelength = 1  # Index of quiescent wavelength in the fitted_parameters
        self.active_wavelength = 5  # Index of active wavelength in the fitted_parameters
        # neural_network
        if self.neural_network is None:
            mlp = MLPClassifier(solver='lbfgs', hidden_layer_sizes=(40,), max_iter=1000,
                                random_state=defaults['random_state'])
            parameter_space = {'alpha': [1e-5, 2e-5, 3e-5, 4e-5, 5e-5, 6e-5, 7e-5, 8e-5, 9e-5]}  # Search region
            # Set GridSearchCV to find best alpha
            self.neural_network = GridSearchCV(mlp, parameter_space, cv=5, n_jobs=-1)

        # STAGE 5: Validate the loaded attributes
        self._validate_attributes()

    def _validate_attributes(self):
        """Validate some of the object's attributes.

        Raises
        ------
        ValueError
            To signal that an attribute is not valid.
        """
        # Arrays must be of length 4
        if len(self.absorption_guess) != 4:
            raise ValueError("absorption_guess should be an array of length 4, got %s" %
                             len(self.absorption_guess))
        if len(self.emission_guess) != 4:
            raise ValueError("emission_guess should be an array of length 4, got %s" %
                             len(self.emission_guess))
        if len(self.absorption_min_bound) != 4:
            raise ValueError("absorption_min_bound should be an array of length 4, got %s" %
                             len(self.absorption_min_bound))
        if len(self.emission_min_bound) != 4:
            raise ValueError("emission_min_bound should be an array of length 4, got %s" %
                             len(self.emission_min_bound))
        if len(self.absorption_max_bound) != 4:
            raise ValueError("absorption_max_bound should be an array of length 4, got %s" %
                             len(self.absorption_max_bound))
        if len(self.emission_max_bound) != 4:
            raise ValueError("emission_max_bound should be an array of length 4, got %s" %
                             len(self.emission_max_bound))
        if len(self.absorption_x_scale) != 4:
            raise ValueError("absorption_x_scale should be an array of length 4, got %s" %
                             len(self.absorption_x_scale))
        if len(self.emission_x_scale) != 4:
            raise ValueError("emission_x_scale should be an array of length 4, got %s" %
                             len(self.emission_x_scale))

        # Absorption and emission fits must be constrained with the correct amplitude sign
        if self.absorption_guess[0] > 0:
            raise ValueError("absorption_guess amplitude should be negative, got %s" %
                             self.absorption_guess[0])
        if self.emission_guess[0] < 0:
            raise ValueError("emission_guess amplitude should be positive, got %s" %
                             self.emission_guess[0])
        if self.emission_min_bound[0] < 0:
            raise ValueError("emission_min_bound amplitude should be positive, got %s" %
                             self.emission_min_bound[0])
        if self.absorption_max_bound[0] > 0:
            raise ValueError("absorption_max_bound amplitude should be negative, got %s" %
                             self.absorption_max_bound[0])

        # Minimum bounds must be less that corresponding maximum bounds
        if np.sum(np.asarray(self.absorption_max_bound) - np.asarray(self.absorption_min_bound) > 0) != \
                len(self.absorption_max_bound):
            raise ValueError("values of absorption_max_bound must be greater than their "
                             "corresponding values in absorption_min_bound")
        if np.sum(np.asarray(self.emission_max_bound) - np.asarray(self.emission_min_bound) > 0) != \
                len(self.emission_max_bound):
            raise ValueError("values of emission_max_bound must be greater than their "
                             "corresponding values in emission_min_bound")

    def _get_sigma(self, classification=None, sigma=None):
        """Infer a sigma profile from the parameters provided.

        If no sigma is provided, use classification to take from the model object's sigma attribute.
        If sigma is provided and is an integer, take from the model object's sigma attribute at that index,
        otherwise, return sigma as a :class:`numpy.ndarray`.

        Parameters
        ----------
        classification : int, optional, default=None
            Classification sigma profile needed to fit.
        sigma : int or array_like, optional, default=None
            Explicit sigma index or profile.

        Returns
        -------
        sigma : numpy.ndarray, length=n_constant_wavelengths
             The sigma profile.

        See Also
        --------
        mcalf.utils.spec.generate_sigma : Generate a specified sigma profile.

        Examples
        --------

        Create a basic model:

        >>> import mcalf.models
        >>> import numpy as np
        >>> wavelengths = np.linspace(8541.3, 8542.7, 30)
        >>> model = mcalf.models.IBIS8542Model(original_wavelengths=wavelengths)

        Choose a sigma profile for the specified classification:

        >>> model._get_sigma(classification=3)
        array([1.        , 1.        , 1.        , 1.        , 1.        ,
               0.99999606, 0.99909053, 0.95597984, 0.55338777, 0.05021681,
               0.05021681, 0.05021681, 0.05021681, 0.4       , 0.4       ,
               0.4       , 0.4       , 0.4       , 0.4       , 0.4       ,
               0.05021681, 0.05021681, 0.05021681, 0.05021681, 0.57661719,
               0.96043995, 0.99922519, 0.99999682, 1.        , 1.        ])

        Get a specific sigma profile "index":

        >>> (model._get_sigma(sigma=1) == model.sigma[1]).all()
        True

        Convert an array like object into a suitable datatype for a sigma profile:

        >>> sigma = [1, 2, 3, 4, 5]
        >>> model._get_sigma(sigma=sigma)
        array([1., 2., 3., 4., 5.])
        """
        if sigma is None:
            # Decide how to weight the wavelengths
            if classification == 0:
                return self.sigma[0]
            else:
                return self.sigma[1]
        else:
            if isinstance(sigma, (int, np.integer)):
                return self.sigma[sigma]
            else:
                return np.asarray(sigma, dtype=np.float64)

    def _fit(self, spectrum, classification=None, spectrum_index=None, profile=None, sigma=None):
        """Fit a single spectrum for the given profile or classification.

        .. warning::
            Using this method directly will skip the corrections that are applied to spectra by the
            :meth:`~mcalf.models.ModelBase.get_spectra` method. Use :meth:`.fit_spectrum` instead.

        Parameters
        ----------
        spectrum : numpy.ndarray, ndim=1, length=n_constant_wavelengths
            The spectrum to be fitted.
        classification : int, optional, default=None
            Classification to determine the fitted profile to use (if profile not explicitly given).
        spectrum_index : array_like or list or tuple, length=3, optional, default=None
            The [time, row, column] index of the `spectrum` provided. Only used for error reporting.
        profile : str, optional, default=None
            The profile to fit. (Will infer profile from classification if omitted.)
        sigma : int or array_like, optional, default=None
            Explicit sigma index or profile. See :meth:`~mcalf.models.IBIS8542Model._get_sigma` for details.

        Returns
        -------
        result : mcalf.models.FitResult
            Outcome of the fit returned in a :class:`~mcalf.models.FitResult` object.

        See Also
        --------
        .fit_spectrum : The recommended method for fitting a single spectrum.
        .fit : The recommended method for fitting multiple spectra.
        mcalf.models.FitResult : The object that the fit method returns.
        """
        if profile is None:  # If profile hasn't been specified, find it

            # Decide which profiles to fit
            if classification in [2, 3, 4]:
                profile = 'both'
            elif classification in [0, 1]:
                profile = 'absorption'
            elif classification is None:
                raise ValueError("classification must be specified if profile is not specified")
            elif not isinstance(classification, (int, np.integer)):
                raise TypeError("classification must be an integer")
            else:
                raise ValueError("unexpected classification, got %s" % classification)

        sigma = self._get_sigma(classification, sigma)

        if profile == 'absorption':
            model = voigt_nobg
            guess = self.absorption_guess
            min_bound = self.absorption_min_bound
            max_bound = self.absorption_max_bound
            x_scale = self.absorption_x_scale
        elif profile == 'emission':
            model = voigt_nobg
            guess = self.emission_guess
            min_bound = self.emission_min_bound
            max_bound = self.emission_max_bound
            x_scale = self.emission_x_scale
        elif profile == 'both':
            model = double_voigt_nobg
            guess = self.absorption_guess + self.emission_guess
            min_bound = self.absorption_min_bound + self.emission_min_bound
            max_bound = self.absorption_max_bound + self.emission_max_bound
            x_scale = self.absorption_x_scale + self.absorption_x_scale
        else:
            raise ValueError("fit profile must be either None, 'absorption', 'emission' or 'both', got %s" % profile)

        time, row, column = spectrum_index if spectrum_index is not None else [None]*3
        fitted_parameters, success = self._curve_fit(model, spectrum, guess, sigma, (min_bound, max_bound), x_scale,
                                                     time=time, row=row, column=column)

        chi2 = np.sum(((spectrum - model(self.constant_wavelengths, *fitted_parameters)) / sigma) ** 2)

        fit_info = {'chi2': chi2, 'classification': classification, 'profile': profile,
                    'success': success, 'index': spectrum_index}

        return FitResult(fitted_parameters, fit_info)

    def plot(self, fit=None, time=None, row=None, column=None, spectrum=None, classification=None, background=None,
             sigma=None, stationary_line_core=None, **kwargs):
        """Plots the data and fitted parameters.

        Parameters
        ----------
        fit : mcalf.models.FitResult or list or array_like, optional, default=None
            The fitted parameters to plot with the data. Can extract the necessary plot metadata from the fit object.
            Otherwise, `fit` should be the parameters to be fitted to either a Voigt or double Voigt profile depending
            on the number of parameters fitted.
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list,
            :class:`numpy.ndarray`, a Python range, etc. can be used. If not provided, will be taken from
            `fit` if it is a :class:`~mcalf.models.FitResult` object, unless a `spectrum` is provided.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectrum : numpy.ndarray, length=`original_wavelengths`, ndim=1, optional, default=None
            The explicit spectrum to plot along with a fit (if specified).
        classification : int, optional, default=None
            Used to determine which sigma profile to use. See :meth:`~mcalf.models.IBIS8542Model._get_sigma`
            for more details. If not provided, will be taken from `fit` if it is a
            :class:`~mcalf.models.FitResult` object, unless a `spectrum` is provided.
        background : float or array_like, length=n_constant_wavelengths, optional, default= see note
            Background to added to the fitted profiles. If a `spectrum` is given, this will default to zero, otherwise
            the value loaded by :meth:`~mcalf.models.ModelBase.load_background` will be used.
        sigma : int or array_like, optional, default=None
            Explicit sigma index or profile. See :meth:`~mcalf.models.IBIS8542Model._get_sigma` for details.
        stationary_line_core : float, optional, default=`stationary_line_core`
            The stationary line core wavelength to mark on the plot.
        **kwargs : dict
            Other parameters used to adjust the plotting.
            See :func:`mcalf.visualisation.plot_ibis8542` for full details.

            * `separate` -- See :meth:`plot_separate`.
            * `subtraction` -- See :meth:`plot_subtraction`.
            * `sigma_scale` -- A factor to multiply the error bars to change their prominence.

        See Also
        --------
        plot_separate : Plot the fit parameters separately.
        plot_subtraction : Plot the spectrum with the emission fit subtracted from it.
        mcalf.models.FitResult.plot : Plotting method on the fit result.

        Examples
        --------
        .. minigallery:: mcalf.visualisation.plot_ibis8542
        """
        if fit.__class__ == FitResult:  # If fit is a `FitResult`

            # Extract classification (so it can choose the correct sigma to plot)
            if classification is None:
                classification = fit.classification

            # Extract spectrum index (so it can load the raw data)
            if (time, row, column) == (None, None, None) and fit.index is not None:
                time, row, column = fit.index

            # Replace fit with just the fitted parameters
            fit = fit.parameters

        # Has a spectrum been given? (used for tidying plot later)
        explicit_spectrum = False if spectrum is None else True

        # Get the spectrum and reduce it to 1D
        spectrum = self.get_spectra(time=time, row=row, column=column, spectrum=spectrum, background=True)
        spectrum = np.squeeze(spectrum)

        # If no background to be added to the fit is given...
        if background is None and fit is not None:
            if not explicit_spectrum:  # Take from loaded background if using indices
                time, row, column = self._get_time_row_column(time=time, row=row, column=column)
                background = self.background[time, row, column]
            else:  # Otherwise assume to be zero
                background = 0

        if classification is not None or sigma is not None:  # Skip if neither classification or sigma are given
            sigma = self._get_sigma(classification, sigma)

        if stationary_line_core is None:
            stationary_line_core = self.stationary_line_core

        return plot_ibis8542(self.constant_wavelengths, spectrum, fit=fit, background=background, sigma=sigma,
                             stationary_line_core=stationary_line_core, **kwargs)

    def plot_separate(self, *args, **kwargs):
        """Plot the fitted profiles separately.

        If multiple profiles exist, fit them separately.
        Arguments are the same as the :meth:`plot` method.

        See Also
        --------
        plot : General plotting method.
        plot_subtraction : Plot the spectrum with the emission fit subtracted from it.
        mcalf.models.FitResult.plot : Plotting method on the fit result.
        """
        self.plot(*args, separate=True, **kwargs)

    def plot_subtraction(self, *args, **kwargs):
        """Plot the spectrum with the emission fit subtracted from it.

        If multiple profiles exist, subtract the fitted emission from the raw data.
        Arguments are the same as the :meth:`plot` method.

        See Also
        --------
        plot : General plotting method.
        plot_separate : Plot the fit parameters separately.
        mcalf.models.FitResult.plot : Plotting method on the fit result.
        """
        self.plot(*args, subtraction=True, **kwargs)


# Copy documentation from base class
IBIS8542_PARAMETERS = copy.deepcopy(BASE_PARAMETERS)
IBIS8542_ATTRIBUTES = copy.deepcopy(BASE_ATTRIBUTES)

# Update documentation from base class (include new defaults)
for d in [IBIS8542_PARAMETERS, IBIS8542_ATTRIBUTES]:
    d['stationary_line_core'] = """
    stationary_line_core : float, optional, default=8542.099145376844
        Wavelength of the stationary line core."""
    d['sigma'] = """
    sigma : list of array_like or bool, length=(2, n_wavelengths), optional, default=[type1, type2]
        A list of different sigma that are used to weight particular wavelengths along the spectra when fitting. The
        fitting method will expect to be able to choose a sigma array from this list at a specific index. It's default
        value is `[generate_sigma(i, constant_wavelengths, stationary_line_core) for i in [1, 2]]`.
        See :func:`mcalf.utils.spec.generate_sigma` for more information.
        If bool, True will generate the default sigma value regardless of the value specified in `config`,
        and False will set `sigma` to be all ones, effectively disabling it."""
IBIS8542_ATTRIBUTES['neural_network'] = """
    neural_network : sklearn.neural_network.MLPClassifier, optional, default= see description
        The :class:`sklearn.neural_network.MLPClassifier` object (or similar) that will be used to
        classify the spectra. Defaults to a :class:`sklearn.model_selection.GridSearchCV`
        with :class:`~sklearn.neural_network.MLPClassifier(solver='lbfgs', hidden_layer_sizes=(40,), max_iter=1000)`
        for best `alpha` selected from `[1e-5, 2e-5, 3e-5, 4e-5, 5e-5, 6e-5, 7e-5, 8e-5, 9e-5]`."""

# Add documentation for new parameters and attributes
IBIS8542_DOCS = """
    absorption_guess : array_like, length=4, optional, default=[-1000, stationary_line_core, 0.2, 0.1]
        Initial guess to take when fitting the absorption Voigt profile.
    emission_guess : array_like, length=4, optional, default=[1000, stationary_line_core, 0.2, 0.1]
        Initial guess to take when fitting the emission Voigt profile.
    absorption_min_bound : array_like, length=4, optional, default=[-np.inf, stationary_line_core-0.15, 1e-6, 1e-6]
        Minimum bounds for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_min_bound : array_like, length=4, optional, default=[0, -np.inf, 1e-6, 1e-6]
        Minimum bounds for all the emission Voigt profile parameters in order of the function's arguments.
    absorption_max_bound : array_like, length=4, optional, default=[0, stationary_line_core+0.15, 1, 1]
        Maximum bounds for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_max_bound : array_like, length=4, optional, default=[np.inf, np.inf, 1, 1]
        Maximum bounds for all the emission Voigt profile parameters in order of the function's arguments.
    absorption_x_scale : array_like, length=4, optional, default=[1500, 0.2, 0.3, 0.5]
        Characteristic scale for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_x_scale : array_like, length=4, optional, default=[1500, 0.2, 0.3, 0.5]
        Characteristic scale for all the emission Voigt profile parameters in order of the function's arguments.
    random_state : int, numpy.random.RandomState, optional, default=None
        Determines random number generation for weights and bias initialisation of the default `neural_network`.
        Pass an int for reproducible results across multiple function calls."""

# Form the docstring and do the replacements
IBIS8542_PARAMETERS_STR = ''.join(IBIS8542_PARAMETERS[i] for i in IBIS8542_PARAMETERS)
IBIS8542_ATTRIBUTES_STR = ''.join(IBIS8542_ATTRIBUTES[i] for i in IBIS8542_ATTRIBUTES)
IBIS8542Model.__doc__ = IBIS8542Model.__doc__.replace(
    '${PARAMETERS}',
    (IBIS8542_DOCS + IBIS8542_PARAMETERS_STR).lstrip()
)
IBIS8542Model.__doc__ = IBIS8542Model.__doc__.replace(
    '${ATTRIBUTES}',
    IBIS8542_DOCS.lstrip()
)
IBIS8542Model.__doc__ = IBIS8542Model.__doc__.replace(
    '${ATTRIBUTES_EXTRA}',
    IBIS8542_ATTRIBUTES_STR.lstrip()
)

update_signature(IBIS8542Model)



import copy

import astropy.units
import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt

from mcalf.utils.plot import calculate_extent

__all__ = ['plot_map']


def plot_map(arr, mask=None, umbra_mask=None, resolution=None, offset=(0, 0), vmin=None, vmax=None,
             lw=None, show_colorbar=True, unit="km/s", ax=None):
    """Plot a velocity map array.

    Parameters
    ----------
    arr : numpy.ndarray[float] or astropy.units.quantity.Quantity, ndim=2
        Two-dimensional array of velocities.
    mask : numpy.ndarray[bool], ndim=2, shape=arr, optional, default=None
        Mask showing the region where velocities were found for. True is outside the
        velocity region and False is where valid velocities should be found. Specifying
        a mask allows for errors in the velocity calculation to be black and points
        outside the region to be gray. If omitted, all invalid points will be gray.
    umbra_mask : numpy.ndarray[bool], ndim=2, shape=arr, optional, default=None
        A mask of the umbra, True outside, False inside. If given, a contour will
        outline the umbra, or other feature the mask represents.
    resolution : tuple[float] or astropy.units.quantity.Quantity, optional, default=None
        A 2-tuple (x, y) containing the length of each pixel in the x and y direction respectively.
        If a value has type :class:`astropy.units.quantity.Quantity`, its axis label will
        include its attached unit, otherwise the unit will default to Mm.
        If `resolution` is None, both axes will be ticked with the default pixel value
        with no axis labels.
    offset : tuple[float] or int, length=2, optional, default=(0, 0)
        Two offset values (x, y) for the x and y axis respectively.
        Number of pixels from the 0 pixel to the first pixel. Defaults to the first
        pixel being at 0 length units. For example, in a 1000 pixel wide dataset,
        setting offset to -500 would place the 0 Mm location at the centre.
    vmin : float, optional, default= ``-max(|arr|)``
        Minimum velocity to plot. If not given, will be -vmax, for vmax not None.
    vmax : float, optional, default= ``max(|arr|)``
        Maximum velocity to plot. If not given, will be -vmin, for vmin not None.
    lw : float, optional, default=None
        The line width of the contour line plotted for `umbra_mask`.
        Passed as `linewidths` to :func:`matplotlib.axes.Axes.contour`.
    show_colorbar : bool, optional, default=True
        Whether to draw a colorbar.
    unit : str or astropy.units.UnitBase or astropy.units.quantity.Quantity, optional, default='km/s'
        The units of `arr` data. Printed on colorbar.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which the velocity map will be plotted.
        Defaults to the current axis of the current figure.

    Returns
    -------
    im : matplotlib.image.AxesImage
        The object returned by :func:`matplotlib.axes.Axes.imshow` after plotting `arr`.

    See Also
    --------
    mcalf.models.FitResults.velocities : Calculate the Doppler velocities for an array of fits.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.plot_map
    """
    if ax is None:
        ax = plt.gca()

    # Validate `arr`
    if not isinstance(arr, np.ndarray) or arr.ndim != 2:
        raise TypeError('`arr` must be a numpy.ndarray with 2 dimensions.')
    arr = arr.copy()  # Edit a copy of `arr`

    # Validate `mask` and `umbra_mask`
    for n, v in (('mask', mask), ('umbra_mask', umbra_mask)):
        if v is not None:
            if not isinstance(v, np.ndarray) or v.ndim != 2:
                raise TypeError(f'`{n}` must be a numpy.ndarray with 2 dimensions.')
            if v.shape != arr.shape:
                raise ValueError(f'`{n}` must be the same shape as `arr`')

    # Update default unit if unit present in `arr`
    if isinstance(arr[0, 0], astropy.units.quantity.Quantity):
        unit = arr.unit.to_string(astropy.units.format.LatexInline)
        arr = arr.value  # Remove unit
    # Convert a `unit` parameter that was provided as an astropy unit
    if isinstance(unit, (astropy.units.UnitBase, astropy.units.quantity.Quantity)):
        unit = unit.to_string(astropy.units.format.LatexInline)

    # Calculate a specific extent if a resolution is specified
    # TODO: Allow the `dimension` to be set by the user.
    extent = calculate_extent(arr.shape, resolution, offset,
                              ax=ax, dimension='distance')

    # Configure default colormap
    cmap = copy.copy(mpl.cm.get_cmap('bwr'))
    cmap.set_bad(color='#999999', alpha=1)

    # Show invalid pixels outside the mask as black, inside as gray
    if mask is not None:

        # Create image from mask
        mask = mask.astype(bool)
        unmasked_section = np.empty_like(mask, dtype=float)
        unmasked_section[mask] = np.nan  # Outside mask
        unmasked_section[~mask] = 1  # Inside mask

        # Configure colormap of mask
        cmap_mask = copy.copy(mpl.cm.get_cmap('gray'))
        cmap_mask.set_bad(color='#999999', alpha=1)

        # Show the masked region
        ax.imshow(unmasked_section, cmap=cmap_mask, origin='lower',
                  extent=extent, interpolation='nearest')

        arr[mask] = np.nan  # Remove values from `arr` that are outside mask
        cmap.set_bad(color='#000000', alpha=0)  # Update default colormap

    # Calculate range for symmetric colormap
    if vmin is None and vmax is None:
        vmax = np.nanmax(np.abs(arr))
        vmin = -vmax
    elif vmin is None:
        vmin = -vmax
    elif vmax is None:
        vmax = -vmin

    # Show the velocities
    im = ax.imshow(arr, cmap=cmap, vmin=vmin, vmax=vmax, origin='lower',
                   extent=extent, interpolation='nearest')

    # Outline the umbra
    if umbra_mask is not None:
        umbra_mask = umbra_mask.astype(bool)
        ax.contour(umbra_mask, [0.5], colors='k', origin='lower',
                   extent=extent, linewidths=lw)

    if show_colorbar:
        ax.get_figure().colorbar(im, ax=[ax], label=f'Doppler velocity ({unit})')

    return im



import numpy as np
import pytest

from mcalf.visualisation import plot_spectrum

from ..helpers import figure_test


@pytest.fixture(scope='function')
def wavelengths_spectrum():
    wavelengths = np.linspace(-1, 1, 30)
    wavelengths = np.delete(wavelengths, [1, 3, 5, 7, -8, -6, -4, -2])
    spectrum = -4.523 * np.exp(-wavelengths ** 2 / 0.05) + 11.354
    wavelengths += 2432.249
    return wavelengths, spectrum


@figure_test
def test_plot_spectrum(pytestconfig, wavelengths_spectrum):
    wavelengths, spectrum = wavelengths_spectrum
    plot_spectrum(wavelengths, spectrum)


@figure_test
def test_plot_spectrum_no_smooth(pytestconfig, wavelengths_spectrum):
    wavelengths, spectrum = wavelengths_spectrum
    plot_spectrum(wavelengths, spectrum, smooth=False)


@figure_test
def test_plot_spectrum_no_norm(pytestconfig, wavelengths_spectrum):
    wavelengths, spectrum = wavelengths_spectrum
    plot_spectrum(wavelengths, spectrum, normalised=False)


@figure_test
def test_plot_spectrum_no_smooth_norm(pytestconfig, wavelengths_spectrum):
    wavelengths, spectrum = wavelengths_spectrum
    plot_spectrum(wavelengths, spectrum, smooth=False, normalised=False)



import os
from pathlib import Path
from functools import wraps

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
import pytest

__all__ = ['data_path_function', 'get_mpl_ft2_version', 'figure_test', 'class_map']


def data_path_function(mod):
    # Returns a function that provides the filename and path of a data file for a particular module
    def data_path(*args, module=mod):
        return pkg_resources.resource_filename('mcalf', os.path.join('tests', module, 'data', *args))
    return data_path


def get_mpl_ft2_version(form):
    """
    Generate the hash library name for this env.
    Parameters
    ----------
    form : {'hash_library', 'baseline_dir'}
    Notes
    -----
    Based on functions at https://github.com/sunpy/sunpy/blob/v2.0.7/sunpy/tests/helpers.py
    The SunPy Community et al. SunPy (Version v2.0.7). Zenodo. http://doi.org/10.5281/zenodo.4423217
    """
    ft2_version = f"{mpl.ft2font.__freetype_version__.replace('.', '')}"
    mpl_version = "dev" if "+" in mpl.__version__ else mpl.__version__.replace('.', '')
    if form == 'hash_library':
        return f"figure_hashes_mpl_{mpl_version}_ft_{ft2_version}.json"
    elif form == 'baseline_dir':
        return f"baseline_mpl_{mpl_version}_ft_{ft2_version}"
    else:
        raise ValueError(f"Unknown `form` {form}")


def figure_test(test_function):
    """
    A decorator for a test that verifies the hash of the current figure or the
    returned figure, with the name of the test function as the hash identifier
    in the library. A PNG is also created in the 'result_image' directory,
    which is created on the current path.
    All such decorated tests are marked with `pytest.mark.figure` for convenient filtering.
    Examples
    --------
    @figure_test
    def test_simple_plot():
        plt.plot([0,1])
    Notes
    -----
    Based on functions at https://github.com/sunpy/sunpy/blob/v2.0.7/sunpy/tests/helpers.py
    The SunPy Community et al. SunPy (Version v2.0.7). Zenodo. http://doi.org/10.5281/zenodo.4423217
    """
    baseline_dir_name = get_mpl_ft2_version('baseline_dir')
    baseline_dir_path = Path(__file__).parent / baseline_dir_name

    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_path,
                                   savefig_kwargs={'metadata': {'Software': None}},
                                   style='default')
    @wraps(test_function)
    def test_wrapper(pytestconfig, *args, **kwargs):
        ret = test_function(pytestconfig, *args, **kwargs)
        if ret is None:
            ret = plt.gcf()
        if pytestconfig.getoption('--mpl', default=None) is None:
            print("close ", end='')
            plt.close(fig=ret)
        else:
            print("return ", end='')
            return ret

    return test_wrapper


def class_map(t=30, x=10, y=20, n=6):
    """Generate a 3D array (t, y, x) of classifications between -1 and n."""
    np.random.seed(4321)
    c = np.random.randint(-1, n, size=(t, y, x))
    c[:, 2, 2] = -1
    return c



import numpy as np

__all__ = ['single_gaussian']


def single_gaussian(x, a, b, c, d):
    """Gaussian function.

    Parameters
    ----------
    x : numpy.ndarray
        Wavelengths to evaluate Gaussian function at.
    a : float
        Amplitude.
    b : float
        Central line core.
    c : float
        Sigma of Gaussian.
    d : float
        Background to add.

    Returns
    -------
    result : numpy.ndarray, shape=`x.shape`
        The value of the Gaussian function here.
    """
    g = a * np.exp(- (x-b)**2.0 / (2.0 * c**2.0))
    return g + d



import astropy.units as u
import numpy as np
import pytest

from mcalf.visualisation import plot_map

from ..helpers import figure_test


@pytest.fixture
def arr():
    np.random.seed(1234)
    a = np.random.rand(8, 6)
    a[3, 3] = np.nan
    a[5, 1] = np.nan
    return (a * 16) - 8


@pytest.fixture
def mask(arr):
    mask = np.full_like(arr, True, dtype=bool)
    mask[1:7, 2:] = False
    return mask


@pytest.fixture
def umbra_mask(arr):
    mask = np.full_like(arr, True, dtype=bool)
    mask[2:6, 2:5] = False
    mask[4, 2] = True
    mask[2, 4] = True
    return mask


@figure_test
def test_plot_map_basic(pytestconfig, arr):
    plot_map(arr)


@figure_test
def test_plot_map_mask(pytestconfig, arr, mask):
    plot_map(arr, mask)


@figure_test
def test_plot_map_umbra_mask(pytestconfig, arr, umbra_mask):
    plot_map(arr, umbra_mask=umbra_mask)


@figure_test
def test_plot_map_both_masks(pytestconfig, arr, mask, umbra_mask):
    plot_map(arr, mask, umbra_mask)


@figure_test
def test_plot_map_resolution(pytestconfig, arr):
    plot_map(arr, resolution=(2.5, 3 * u.kg / u.Hz))


@figure_test
def test_plot_map_resolution_offset(pytestconfig, arr):
    plot_map(arr, resolution=(2.5, 3 * u.kg / u.Hz), offset=(-3, -4))


@figure_test
def test_plot_map_resolution_offset_masks(pytestconfig, arr, mask, umbra_mask):
    plot_map(arr, mask, umbra_mask, resolution=(2.5, 3 * u.kg / u.Hz), offset=(-3, -4))


@figure_test
def test_plot_map_vmin(pytestconfig, arr):
    plot_map(arr, vmin=-4.5)


@figure_test
def test_plot_map_vmax(pytestconfig, arr):
    plot_map(arr, vmax=4.5)


@figure_test
def test_plot_map_unit(pytestconfig, arr):
    plot_map(arr, unit='test/unit')


@figure_test
def test_plot_map_unit_astropy(pytestconfig, arr):
    plot_map(arr, unit=(u.m / u.s))


@figure_test
def test_plot_map_unit_arr_astropy(pytestconfig, arr):
    plot_map(arr * u.m / u.s, unit='test/unit')


@figure_test
def test_plot_map_lw(pytestconfig, arr, umbra_mask):
    plot_map(arr, umbra_mask=umbra_mask, lw=5.)


@figure_test
def test_plot_map_colorbar(pytestconfig, arr):
    plot_map(arr, show_colorbar=False)


def test_plot_map_validate_arr(arr):
    for a in ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], np.array([arr, arr])):
        with pytest.raises(TypeError) as e:
            plot_map(a)
        assert '`arr` must be a numpy.ndarray with 2 dimensions' in str(e.value)


def test_plot_map_validate_masks(arr, mask, umbra_mask):

    # Mask is 2D array
    for a in ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], np.array([mask, mask])):
        with pytest.raises(TypeError) as e:
            plot_map(arr, a, umbra_mask)
        assert '`mask` must be a numpy.ndarray with 2 dimensions' in str(e.value)

    # Umbra mask is 2D array
    for a in ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], np.array([umbra_mask, umbra_mask])):
        with pytest.raises(TypeError) as e:
            plot_map(arr, mask, a)
        assert '`umbra_mask` must be a numpy.ndarray with 2 dimensions' in str(e.value)

    # Mask wrong shape
    with pytest.raises(ValueError) as e:
        plot_map(arr, np.vstack([mask, mask]), umbra_mask)
    assert '`mask` must be the same shape as `arr`' in str(e.value)

    # Umbra mask wrong shape
    with pytest.raises(ValueError) as e:
        plot_map(arr, mask, np.vstack([umbra_mask, umbra_mask]))
    assert '`umbra_mask` must be the same shape as `arr`' in str(e.value)



import matplotlib.pyplot as plt
import numpy as np

from mcalf.profiles.voigt import double_voigt, voigt
from mcalf.utils.plot import hide_existing_labels
from mcalf.utils.spec import reinterpolate_spectrum

__all__ = ['plot_ibis8542', 'plot_spectrum']


def plot_ibis8542(wavelengths, spectrum, fit=None, background=0,
                  sigma=None, sigma_scale=70,
                  stationary_line_core=None,
                  subtraction=False, separate=False,
                  show_intensity=True, show_legend=True, ax=None):
    """Plot an :class:`~mcalf.models.IBIS8542Model` fit.

    .. note::
        It is recommended to use the plot method built into either the
        :class:`~mcalf.models.IBIS8542Model` class or the :class:`~mcalf.models.FitResult`
        class instead.

    Parameters
    ----------
    wavelengths : numpy.ndarray
        The x-axis values.
    spectrum : numpy.ndarray, length=n_wavelengths
        The y-axis values.
    fit : array_like, optional, default=None
        The fitted parameters.
    background : float or numpy.ndarray, length=n_wavelengths, optional, default=0
        The background to add to the fitted profiles.
    sigma : numpy.ndarray, length=n_wavelengths, optional, default=None
        The sigma profile used when fitting the parameters to `spectrum`.
        If given, will be plotted as shaded regions.
    sigma_scale : float, optional, default=70
        A factor to multiply the error bars to change their prominence.
    stationary_line_core : float, optional, default=None
        If given, will show a dashed line at this wavelength.
    subtraction : bool, optional, default=False
        Whether to plot the `spectrum` minus emission fit (if exists) instead.
    separate : bool, optional, default=False
        Whether to plot the fitted profiles separately (if multiple components exist).
    show_intensity : bool, optional, default=True
        Whether to show the intensity axis tick labels and axis label.
    show_legend : bool, optional, default=True
        Whether to draw a legend on the axes.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which the fit will be plotted.
        Defaults to the current axis of the current figure.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axes the lines are drawn on.

    See Also
    --------
    mcalf.models.IBIS8542Model.plot : General plotting method.
    mcalf.models.IBIS8542Model.plot_separate : Plot the fit parameters separately.
    mcalf.models.IBIS8542Model.plot_subtraction : Plot the spectrum with the emission fit subtracted from it.
    mcalf.models.FitResult.plot : Plotting method provided by the fit result.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.plot_ibis8542
    """
    if ax is None:
        ax = plt.gca()

    plot_settings = {
        'obs': {'color': '#006BA4', 'label': 'observation'},
        'abs': {'color': '#A2C8EC', 'label': 'absorption profile'},
        'emi': {'color': '#595959', 'label': 'emission profile'},
        'fit': {'color': '#FF800E', 'label': 'fitted profile'},
        'alc': {'color': '#A2C8EC', 'label': 'absorption line core', 'linestyle': '--'},
        'elc': {'color': '#595959', 'label': 'emission line core', 'linestyle': ':'},
        'slc': {'color': '#ABABAB', 'label': 'stationary line core', 'linestyle': '--'},
    }
    hide_existing_labels(plot_settings, fig=ax.get_figure())

    if fit is None:
        show_fit = show_sigma = False
        subtraction = separate = False  # not possible, ignore request
    else:  # fitted parameters provided
        show_fit = True
        show_sigma = False if sigma is None else True
        if len(fit) == 8:  # two components fitted
            fit_function = double_voigt
        else:  # one component fitted
            fit_function = voigt
            subtraction = separate = False  # not possible, ignore request

    if subtraction:
        spectrum = spectrum - voigt(wavelengths, *fit[4:], 0, clib=False)
        plot_settings['obs']['label'] = 'observation - emission'
        show_fit = show_sigma = None

    if show_sigma:  # make a shaded region around the spectral data
        ax.fill_between(wavelengths,
                        spectrum-sigma*sigma_scale, spectrum+sigma*sigma_scale,
                        color='lightgrey')

    # Plot the spectral data
    ax.plot(wavelengths, spectrum, **plot_settings['obs'])

    # Plot the fitted profiles if parameters are given
    if show_fit:

        if separate:  # Plot each component separately

            ax.plot(wavelengths,
                    double_voigt(wavelengths, *fit, background, clib=False),
                    **plot_settings['fit'])
            ax.plot(wavelengths,
                    voigt(wavelengths, *fit[:4], background, clib=False),
                    **plot_settings['abs'])
            ax.plot(wavelengths,
                    voigt(wavelengths, *fit[4:], 0, clib=False),
                    **plot_settings['emi'])

        else:  # Plot a combined profile

            ax.plot(wavelengths,
                    fit_function(wavelengths, *fit, background, clib=False),
                    **plot_settings['fit'])

        # Plot absorption line core
        ax.axvline(x=fit[1], **plot_settings['alc'])

        # Plot emission line core (if fitted)
        if fit_function is double_voigt:
            ax.axvline(x=fit[5], **plot_settings['elc'])

    # Plot stationary line core
    if stationary_line_core is not None:  # Plot vertical dashed line
        ax.axvline(x=stationary_line_core, **plot_settings['slc'])

    # Format the axis ticks and labels:
    ax.minorticks_on()
    ax.set_xlabel('wavelength (Å)')
    if show_intensity:
        ax.set_ylabel('intensity')
    else:  # hide y-axis and axes box
        for loc, spine in ax.spines.items():
            if loc != 'bottom':
                spine.set_color('none')  # don't draw spine
            ax.yaxis.set_ticks([])
            ax.xaxis.set_ticks_position('bottom')

    if show_legend:
        bbox = ax.get_position()
        ax.set_position([bbox.x0, bbox.y0, bbox.width * 0.6, bbox.height])
        ax.legend(loc='upper center', bbox_to_anchor=(1.45, 0.8), ncol=1)

    return ax


def plot_spectrum(wavelengths, spectrum, normalised=True, smooth=True, ax=None):
    """Plot a spectrum with the wavelength grid shown.

    Intended for plotting the raw data.

    Parameters
    ----------
    wavelengths : numpy.ndarray
        The x-axis values.
    spectrum : numpy.ndarray, length=n_wavelengths
        The y-axis values.
    normalised : bool, optional, default=True
        Whether to normalise the spectrum using the last three spectral points.
    smooth : bool, optional, default=True
        Whether to smooth the `spectrum` with a spline.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which the fit will be plotted.
        Defaults to the current axis of the current figure.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axes the lines are drawn on.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.plot_spectrum
    """
    # Inputted wavelengths and spectrum
    wavelengths_pts = wavelengths.copy()
    spectrum_pts = spectrum.copy()

    if smooth:
        wavelengths = np.linspace(wavelengths[0], wavelengths[-1], num=200)
        spectrum = reinterpolate_spectrum(spectrum, wavelengths_pts, wavelengths)

    if normalised:
        norm = np.mean(spectrum_pts[-3:])
        spectrum_pts /= norm
        spectrum /= norm

    if ax is None:
        ax = plt.gca()

    # Plot vertical lines at each wavelength
    for x in wavelengths_pts:
        ax.axvline(x=x, linestyle='--', linewidth=0.5, color='black')

    # Plot spectral line
    ax.plot(wavelengths, spectrum, color='black')

    # Plot markers at wavelength points
    ax.plot(wavelengths_pts, spectrum_pts, marker='o', linestyle='None', color='black', markersize=4)

    # Set axis labels and ticks
    ax.set_xlabel('wavelength (Å)')
    ylabel = 'normalised intensity ($I/I_c$)' if normalised else 'intensity ($I$)'
    ax.set_ylabel(ylabel)
    ax.minorticks_on()
    for label in ax.xaxis.get_ticklabels()[::2]:
        label.set_visible(False)

    return ax



import copy
import collections

import numpy as np
import pytest

from mcalf.utils.collections import OrderedParameterDict, Parameter, ParameterDict


def test_parameter():

    apples = Parameter('apples')
    oranges = Parameter('oranges', 12)
    Parameter('bananas', value=15)
    Parameter(name='grapes', value=30)

    # No value set

    assert apples == 'apples'
    assert apples is not None
    assert str(apples) == 'apples'
    assert apples.name == 'apples'
    assert apples.value is None
    assert apples.value_or_name == 'apples'
    assert apples == apples
    assert apples + 1 == apples + 1

    assert apples + 7 == 'apples+7'
    assert apples - 8 == 'apples-8'
    assert apples * 3 == 'apples*3'
    assert apples / 11 == 'apples/11'
    assert apples * 12 + 3 == 'apples*12+3'
    assert apples / 6 - 6 / 2 == 'apples/6-3.0'

    with pytest.raises(ValueError) as e:
        apples.eval()
    assert 'Cannot evaluate without a value.' in str(e.value)

    # Value set

    assert oranges == 12
    assert not oranges == '12'
    assert not oranges == 'oranges'
    assert str(oranges) == '12'
    assert oranges.name == 'oranges'
    assert oranges.value == 12
    assert oranges.value_or_name == 12
    assert oranges == oranges
    assert oranges + 1 == oranges + 1

    assert oranges + 2 == 14
    assert str(oranges + 2) == 'oranges+2'
    assert repr(oranges + 2) == "'oranges+2'"
    assert (oranges * 10 + 4).eval() == 124

    assert Parameter('a', 1) == Parameter('b', 1)
    assert Parameter('a', 1) + 0 == Parameter('b', 1) + 0

    # Equations that need brackets to work are not supported
    with pytest.raises(NotImplementedError):
        (oranges + 2) * 3

    # Parameter must be left-most object
    with pytest.raises(TypeError):
        2 + oranges
    with pytest.raises(TypeError):
        6.0 * oranges
    with pytest.raises(TypeError):
        2 + oranges * 2

    # Parameter + Parameter evaluates RHS parameter
    # (This functionality should not be considered stable.)
    assert str(Parameter('a') * 2 + Parameter('b', 6) / 2) == 'a*2+3.0'

    # Copying a parameter
    a = Parameter('A', value=1)
    b = a.copy()
    c = copy.copy(a)
    d = copy.deepcopy(a)
    e = a + 0
    f = a
    a.value = 2
    for p in (b, c, d, e):
        assert p == 1
    assert f == 2

    # Very basic array support
    a = Parameter('a', np.arange(10))
    assert np.array_equal((a + 1).eval(), np.arange(10) + 1)
    with pytest.raises(SyntaxError):
        np.array_equal((a + 2 * np.arange(10)).eval(), 3 * np.arange(10))


def verify_synced_parameters(tracked):
    for p in tracked.values():
        for obj in p['objects']:
            assert obj.value is p['value']


@pytest.mark.parametrize('cls', [OrderedParameterDict, ParameterDict])
def test_parameter_dict(cls):

    # Initialise with parameters
    parameters = [
        ('aa', Parameter('a', 2) + 1),
        ('bb', Parameter('b') + 2),
        ('cc', Parameter('a', 2) + 3),
        ('dd', Parameter('c') + 4),
        ('ee', Parameter('b', 3) + 5),
        ('ff', Parameter('f') + 6.25),
        ('gg', Parameter('f', 13.2) + 3.12),
        ('hh', [Parameter('f') + 1, Parameter('b'), 12]),
    ]

    if cls is ParameterDict:
        parameters = {k: v for k, v in parameters}
    x = cls(parameters)

    assert len(x._tracked.keys()) == 4
    assert set(x._tracked.keys()) == {'a', 'b', 'c', 'f'}

    verify_synced_parameters(x._tracked)
    assert x['aa'] == 3
    assert x['bb'] == 5
    assert x['cc'] == 5
    assert x['dd'] == 'c+4'
    assert x['ee'] == 8
    assert x['ff'] == 19.45
    assert x['gg'] == 16.32
    assert isinstance(x['hh'], list) and x['hh'] == [14.2, 3, 12]
    assert x.has_value('a')
    assert not x.has_value('c')
    assert x.get_parameter('c') is None

    # Update some parameters
    x.update_parameter('c', 9)
    x.update_parameter('b', None)
    x.update_parameter('a', 12)
    x.update_parameter('f', 5.8)

    verify_synced_parameters(x._tracked)
    assert x['aa'] == 13
    assert x['bb'] == 'b+2'
    assert x['cc'] == 15
    assert x['dd'] == 13
    assert x['ee'] == 'b+5'
    assert x['ff'] == 12.05
    assert x['gg'] == 8.92
    assert isinstance(x['hh'], list) and x['hh'] == [6.8, 'b', 12]

    # Evaluate all parameters
    with pytest.raises(ValueError):
        x.eval()
    x.update_parameter('b', 1)
    ex = x.eval()
    if cls is ParameterDict:
        assert isinstance(ex, collections.UserDict)
    elif cls is OrderedParameterDict:
        assert isinstance(ex, collections.OrderedDict)
    for k, v in [('aa', 13), ('bb', 3), ('cc', 15), ('dd', 13), ('ee', 6), ('ff', 12.05), ('gg', 8.92)]:
        if isinstance(v, float):
            assert isinstance(ex[k], float) and ex[k] == v
        else:
            assert ex[k] is v
            assert not x[k] is v

    assert x.exists('b')
    assert not x.exists('t')

    # Add parameter later
    x['tt'] = Parameter('t', 22) + 5
    assert x.exists('t')
    assert x['tt'] == 27
    assert x.get_parameter('t') == 22
    verify_synced_parameters(x._tracked)
    assert len(x._tracked.keys()) == 5
    assert set(x._tracked.keys()) == {'a', 'b', 'c', 'f', 't'}

    # Add standard number
    x['s'] = 10
    assert x['s'] == 10

    # Add parameter with different value
    with pytest.raises(ValueError) as e:
        x['uu'] = Parameter('t', 20) + 7
    assert 'does not match' in str(e.value)



import os

import matplotlib.pyplot as plt
import numpy as np
import pytest
from astropy.io import fits
from sklearn.datasets import make_classification
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import cross_val_score

from mcalf.models import FitResults, IBIS8542Model, ModelBase
from mcalf.profiles.voigt import double_voigt, voigt

from ..helpers import data_path_function, figure_test

data_path = data_path_function('models')


class DummyClassifier:

    def __init__(self, trained=False, n_features=None, classifications=None):
        self.trained = trained
        self.n_features = n_features
        self.defined_return = True if classifications is not None else False
        self.classifications = classifications

    def train(self, X, y):
        assert np.ndim(X) == 2  # (n_samples, n_features)
        assert np.ndim(y) == 1  # (n_samples)
        assert len(X) == len(y)
        self.trained = True
        self.n_features = X.shape[-1]

    def test(self, X, y):
        if not self.trained:
            raise NotFittedError()
        assert np.ndim(X) == 2  # (n_samples, n_features)
        assert np.ndim(y) == 1  # (n_samples)
        assert X.shape[-1] == self.n_features
        assert len(X) == len(y)

    def predict(self, X):
        if self.defined_return:
            return self.classifications.flatten()
        if not self.trained:
            raise NotFittedError()
        assert np.ndim(X) == 2  # (n_samples, n_features)
        assert X.shape[-1] == self.n_features
        return np.asarray(X[:, -1] * 100, dtype=int)


@pytest.mark.parametrize('model', (ModelBase, IBIS8542Model))
def test_default_parameters(model):
    # Test default parameters
    with pytest.raises(ValueError) as e:
        model()
    assert 'original_wavelengths' in str(e.value)


def test_ibis8542model_basic():
    # Will break if default parameters are changes in mcalf.models.IBIS8542Model
    wl = 1000.97
    x_orig = np.linspace(999.81, 1002.13, num=25)
    prefilter_main = 1 - np.abs(x_orig - wl) * 0.1
    prefilter_wvscl = x_orig - wl
    m = IBIS8542Model(stationary_line_core=wl, original_wavelengths=x_orig,
                      prefilter_ref_main=prefilter_main, prefilter_ref_wvscl=prefilter_wvscl)

    bg = 1327.243
    arr = voigt(x_orig, -231.42, wl+0.05, 0.2, 0.21, bg)

    m.load_array(np.array([arr, arr]), names=['row', 'wavelength'])
    m.load_background(np.array([bg, bg]), names=['row'])

    # Fit with assumed neural network classification
    fit1, fit2 = m.fit(row=[0, 1], classifications=np.array([0, 0]))
    fit3 = m.fit_spectrum(arr, classifications=0, background=bg)  # Make sure this is equiv
    truth = [-215.08275199, 1001.01035476, 0.00477063, 0.28869302]

    assert np.array_equal(fit1.parameters, fit2.parameters)
    assert np.array_equal(fit1.parameters, fit3.parameters)
    assert list(fit1.parameters) == pytest.approx(truth)


def test_ibis8542model_configfile():
    # Enter the data directory (needed to find the files referenced in the config files)
    original_dir = os.getcwd()
    os.chdir(data_path())

    # Test with config file
    IBIS8542Model(config="ibis8542model_config.yml")

    # Turn off sigma
    IBIS8542Model(config="ibis8542model_config.yml", sigma=False)

    # Test with defined prefilter
    with pytest.deprecated_call():
        IBIS8542Model(config="ibis8542model_config_prefilter.yml")

    # Test with no prefilter
    IBIS8542Model(config="ibis8542model_config_noprefilter.yml")

    # TODO Check that the parameters were imported correctly

    # Go back to original directory
    os.chdir(original_dir)


@pytest.fixture
def valid_kwargs():
    stationary_line_core = 8542.099145376844
    defaults = {
        'stationary_line_core': stationary_line_core,
        'original_wavelengths': np.linspace(stationary_line_core - 2, stationary_line_core + 2, num=25),
        'constant_wavelengths': np.linspace(stationary_line_core - 1.7, stationary_line_core + 1.7, num=30),
        'prefilter_response': np.loadtxt(data_path('ibis8542model_prefilter.csv'), delimiter=','),
    }
    return defaults


@pytest.mark.parametrize('model', (ModelBase, IBIS8542Model))
def test_validate_parameters(model, valid_kwargs):

    # Default parameters should work
    model(**valid_kwargs)

    # original_wavelengths or constant_wavelengths not sorted
    for wavelengths in ('original_wavelengths', 'constant_wavelengths'):
        with pytest.raises(ValueError) as e:
            defaults_mod = valid_kwargs.copy()
            unsorted = defaults_mod[wavelengths].copy()  # Copy the numpy array before shuffling
            unsorted[10] = unsorted[12]  # Copy the 12th element to the 10th (now out of order)
            defaults_mod[wavelengths] = unsorted  # Replace the sorted array with the unsorted
            model(**defaults_mod)  # Initialise the model
        assert wavelengths in str(e.value) and 'must be sorted ascending' in str(e.value)

    # assert warning when constant_wavelengths extrapolates original_wavelengths (1e-5 over below and above)
    delta_lambda = 0.05
    for match, i, delta in [
        ("Upper bound of `constant_wavelengths` is outside of `original_wavelengths` range.", -1, delta_lambda + 1e-5),
        ("Lower bound of `constant_wavelengths` is outside of `original_wavelengths` range.", 0, -1e-5)
    ]:
        with pytest.warns(Warning, match=match):
            defaults_mod = valid_kwargs.copy()
            extrapolate = defaults_mod['constant_wavelengths'].copy()
            extrapolate[i] = defaults_mod['original_wavelengths'][i] + delta
            defaults_mod['constant_wavelengths'] = extrapolate
            model(**defaults_mod, delta_lambda=delta_lambda)  # Initialise the model

    # stationary_line_core is not a float
    with pytest.raises(ValueError) as e:
        defaults_mod = valid_kwargs.copy()
        defaults_mod.update({'stationary_line_core': int(8542)})
        model(**defaults_mod)
    assert 'stationary_line_core' in str(e.value) and 'float' in str(e.value)

    # stationary_line_core out of wavelength range
    with pytest.raises(ValueError) as e:
        defaults_mod = valid_kwargs.copy()
        defaults_mod.update({'stationary_line_core': float(1000)})
        model(**defaults_mod)
    assert 'stationary_line_core' in str(e.value) and 'is not within' in str(e.value)

    # Check that length of prefilter response is enforced
    with pytest.raises(ValueError) as e:
        defaults_mod = valid_kwargs.copy()
        defaults_mod.update({'prefilter_response': valid_kwargs['prefilter_response'][:-1]})
        m = model(**defaults_mod)
        m._set_prefilter()
    assert 'prefilter_response' in str(e.value) and 'must be the same length' in str(e.value)

    # Check that unexpected kwargs raise an error
    with pytest.raises(TypeError) as e:
        model(**valid_kwargs, qheysnfebsy=None)
    assert 'got an unexpected keyword argument' in str(e.value) and 'qheysnfebsy' in str(e.value)

    # TODO Verify remaining conditions


@pytest.mark.parametrize('model', (ModelBase, IBIS8542Model))
def test_load_data(model, valid_kwargs):

    # Initialise model
    m = model(**valid_kwargs)

    # Unknown target
    array = np.random.rand(5, 10, 15, 30)
    with pytest.raises(ValueError) as e:
        m._load_data(array, names=['time', 'row', 'column', 'wavelength'], target='arrrrray')
    assert 'array target must be' in str(e.value) and 'arrrrray' in str(e.value)

    # Ensure a single spectrum array cannot be loaded (not supported)
    for target, names, array in [('array', ['wavelength'], np.random.rand(30)), ('background', [], 723.23)]:
        with pytest.raises(ValueError) as e:
            m._load_data(array, names=names, target=target)
        assert 'cannot load an array containing one spectrum' in str(e.value)

    # Ensure dimension names are validated
    for meth in (m.load_array, m.load_background):

        array = np.random.rand(5, 10, 15, 30)
        with pytest.raises(ValueError) as e:
            meth(array)
        assert 'dimension names must be specified' in str(e.value)

        array = np.random.rand(5, 10, 15, 30)
        with pytest.raises(ValueError) as e:
            meth(array, names=['row', 'column', 'wavelength'])
        assert 'number of dimension names do not match number of columns' in str(e.value)

        array = np.random.rand(5, 10, 15, 30)
        with pytest.raises(ValueError) as e:
            meth(array, names=['wavelength', 'row', 'column', 'wavelength'])
        assert 'duplicate dimension names found' in str(e.value)

    array = np.random.rand(5, 10, 15)
    with pytest.raises(ValueError) as e:
        m.load_array(array, names=['time', 'row', 'column'])
    assert 'array must contain a wavelength dimension' in str(e.value)

    # 'wavelengths' not valid background dimension
    array = np.random.rand(5, 10, 15)
    with pytest.raises(ValueError) as e:
        m.load_background(array, names=['row', 'column', 'wavelength'])
    assert "name 'wavelength' is not a valid dimension name" in str(e.value)

    # 'rows' should be 'row' etc.
    array = np.random.rand(5, 10, 15, 30)
    with pytest.raises(ValueError) as e:
        m.load_array(array, names=['time', 'rows', 'column', 'wavelength'])
    assert "name 'rows' is not a valid dimension name" in str(e.value)
    array = np.random.rand(5, 10, 15)
    with pytest.raises(ValueError) as e:
        m.load_background(array, names=['time', 'rows', 'column'])
    assert "name 'rows' is not a valid dimension name" in str(e.value)

    array = np.random.rand(5, 10, 15, 30)
    with pytest.raises(ValueError) as e:
        m.load_array(array, names=['time', 'row', 'column', 'wavelength'])
    assert 'length of wavelength dimension not equal length of original_wavelengths' in str(e.value)

    array = np.random.rand(5, 10, 15)
    m.load_background(array, names=['time', 'row', 'column'])
    array = np.random.rand(5, 7, 15, 25)
    with pytest.warns(UserWarning, match="shape of spectrum array indices does not match shape of background array"):
        m.load_array(array, names=['time', 'row', 'column', 'wavelength'])


@pytest.mark.parametrize('model', (ModelBase, IBIS8542Model))
def test_get_time_row_column(model, valid_kwargs):

    # Initialise model
    m = model(**valid_kwargs)
    array = np.random.rand(4, 5, 6, 25)
    names = ['time', 'row', 'column', 'wavelength']
    m.load_array(array.copy(), names=names)

    # (test get_spectra with multiple spectra -- only supports a single spectrum)
    with pytest.raises(ValueError) as e:
        m.get_spectra(spectrum=np.random.rand(10, 30))
    assert 'explicit spectrum must have one dimension' in str(e.value)

    # All dimensions loaded so make sure all dimensions required
    for dim in names[:-1]:
        with pytest.raises(ValueError) as e:
            drop_a_name = names[:-1].copy()
            drop_a_name.remove(dim)
            m._get_time_row_column(**{k: 2 for k in drop_a_name})
        assert f"{dim} index must be specified as multiple indices exist" == str(e.value)

    # Drop a dimension and don't specify it
    array = array[0]  # Drop one (arbitrary) dimension
    for dim in names[:-1]:
        drop_a_name = names.copy()
        drop_a_name.remove(dim)
        m.load_array(array.copy(), names=drop_a_name)
        m._get_time_row_column(**{k: 2 for k in drop_a_name[:-1]})


def test_modelbase_fit(valid_kwargs):
    m = ModelBase(**valid_kwargs)
    array = np.random.rand(4, 5, 6, 25)
    array[0] = np.nan  # For testing no spectra given
    names = ['time', 'row', 'column', 'wavelength']
    m.load_array(array.copy(), names=names)

    # `_fit` method must be implemented in a subclass
    with pytest.raises(NotImplementedError):
        m._fit(np.random.rand(30))

    # Must raise exception if no spectra presented for fitting
    with pytest.raises(ValueError) as e:
        m.fit(time=0, row=range(5), column=range(6))
    assert 'no valid spectra given' in str(e.value)


def test_ibis8542model_validate_parameters(valid_kwargs):

    stationary_line_core = valid_kwargs['stationary_line_core']
    defaults = {
        'absorption_guess': [-1000, stationary_line_core, 0.2, 0.1],
        'emission_guess': [1000, stationary_line_core, 0.2, 0.1],
        'absorption_min_bound': [-np.inf, stationary_line_core - 0.15, 1e-6, 1e-6],
        'emission_min_bound': [0, -np.inf, 1e-6, 1e-6],
        'absorption_max_bound': [0, stationary_line_core + 0.15, 1, 1],
        'emission_max_bound': [np.inf, np.inf, 1, 1],
        'absorption_x_scale': [1500, 0.2, 0.3, 0.5],
        'emission_x_scale': [1500, 0.2, 0.3, 0.5]
    }

    # Incorrect Lengths
    for key, value in defaults.items():  # For each parameter
        with pytest.raises(ValueError) as e:  # Catch ValueErrors
            defaults_mod = defaults.copy()  # Create a copy of the default parameters
            defaults_mod.update({key: value[:-1]})  # Crop the parameter's value
            IBIS8542Model(**valid_kwargs, **defaults_mod)  # Pass the cropped parameter with the other default params
        assert key in str(e.value) and 'length' in str(e.value)  # Error must be about length of current parameter

    # Check that the sign of the following amplitudes are enforced
    for sign, bad_number, parameters in [('positive', -10.42, ('emission_guess', 'emission_min_bound')),
                                         ('negative', +10.42, ('absorption_guess', 'absorption_max_bound'))]:
        for p in parameters:
            with pytest.raises(ValueError) as e:
                bad_value = defaults[p].copy()
                bad_value[0] = bad_number
                defaults_mod = defaults.copy()
                defaults_mod.update({p: bad_value})
                IBIS8542Model(**valid_kwargs, **defaults_mod)
            assert p in str(e.value) and sign in str(e.value)

    # TODO Verify remaining conditions


def test_ibis8542model_get_sigma():
    # Enter the data directory (needed to find the files referenced in the config files)
    original_dir = os.getcwd()
    os.chdir(data_path())

    m = IBIS8542Model(config="ibis8542model_config.yml")
    sigma = np.loadtxt("ibis8542model_sigma.csv", delimiter=',')

    assert np.array_equal(m._get_sigma(classification=0), sigma[0])

    for c in (1, 2, 3, 4):
        assert np.array_equal(m._get_sigma(classification=c), sigma[1])

    for i in (0, 1):
        assert np.array_equal(m._get_sigma(sigma=i), sigma[i])

    x = np.array([1.4325, 1421.43, -1325.342, 153.3, 1.2, 433.0])
    assert np.array_equal(m._get_sigma(sigma=x), x)

    # Go back to original directory
    os.chdir(original_dir)


@pytest.fixture(scope='module')
def ibis8542model_init():
    # Enter the data directory (needed to find the files referenced in the config files)
    original_dir = os.getcwd()
    os.chdir(data_path())

    x_orig = np.loadtxt("ibis8542model_wavelengths_original.csv")

    m = IBIS8542Model(config="ibis8542model_init_config.yml", constant_wavelengths=x_orig, original_wavelengths=x_orig,
                      prefilter_response=np.ones(25))

    m.neural_network = DummyClassifier(trained=True, n_features=25)

    # Go back to original directory
    os.chdir(original_dir)

    return m


def test_ibis8542model_classify_spectra(ibis8542model_init):

    np.random.seed(0)
    spectra = np.random.rand(30, 25) * 100 + 1000
    truth = np.array([10, 37, 72, 0, 27, 100, 81, 44, 17, 49, 22, 97, 28, 100, 22, 17,
                      64, 91, 94, 47, 0, 34, 96, 98, 97, 3, 82, 87, 40, 51], dtype=int)

    m = ibis8542model_init

    classifications = m.classify_spectra(spectra=spectra, only_normalise=True)
    assert np.array_equal(classifications, truth)

    for i in range(len(truth)):
        classifications = m.classify_spectra(spectra=spectra[i], only_normalise=False)
        assert classifications[0] == truth[i] and len(classifications) == 1

    m.neural_network.trained = False
    with pytest.raises(NotFittedError):
        m.classify_spectra(spectra=spectra, only_normalise=True)


@pytest.fixture(scope='module')
def ibis8542model_spectra(ibis8542model_init):
    """IBIS8542Model with random data loaded"""

    m = ibis8542model_init

    spectra = np.empty((2, 3, 4, len(m.original_wavelengths)), dtype=np.float64)
    classifications = np.empty((2, 3, 4), dtype=int)

    np.random.seed(253)
    a1_array = np.random.rand(*spectra.shape[:-1]) * 500 - 800
    a2_array = np.random.rand(*spectra.shape[:-1]) * 500 + 300
    b1_array = np.random.rand(*spectra.shape[:-1]) / 2 - 0.25 + m.stationary_line_core
    b2_array = np.random.rand(*spectra.shape[:-1]) / 2 - 0.25 + m.stationary_line_core
    s1_array = np.random.rand(*spectra.shape[:-1]) / 2 + 0.1
    s2_array = np.random.rand(*spectra.shape[:-1]) / 2 + 0.1
    g1_array = np.random.rand(*spectra.shape[:-1]) / 2 + 0.1
    g2_array = np.random.rand(*spectra.shape[:-1]) / 2 + 0.1
    d_array = np.random.rand(*spectra.shape[:-1]) * 600 + 700

    a1_array[0, 1] = np.nan
    a2_array[1, :2] = np.nan

    for i in range(len(spectra)):
        for j in range(len(spectra[0])):
            for k in range(len(spectra[0, 0])):
                if np.isnan(a1_array[i, j, k]):
                    classifications[i, j, k] = 4
                    spectra[i, j, k] = voigt(m.original_wavelengths, a2_array[i, j, k], b2_array[i, j, k],
                                             s2_array[i, j, k], g2_array[i, j, k], d_array[i, j, k])
                elif np.isnan(a2_array[i, j, k]):
                    classifications[i, j, k] = 0
                    spectra[i, j, k] = voigt(m.original_wavelengths, a1_array[i, j, k], b1_array[i, j, k],
                                             s1_array[i, j, k], g1_array[i, j, k], d_array[i, j, k])
                else:
                    classifications[i, j, k] = 1
                    spectra[i, j, k] = double_voigt(m.original_wavelengths, a1_array[i, j, k], a2_array[i, j, k],
                                                    b1_array[i, j, k], b2_array[i, j, k],
                                                    s1_array[i, j, k], s2_array[i, j, k],
                                                    g1_array[i, j, k], g2_array[i, j, k], d_array[i, j, k])

    m.load_array(spectra, names=['time', 'row', 'column', 'wavelength'])
    m.load_background(d_array, names=['time', 'row', 'column'])

    return m, classifications


@pytest.fixture(scope='module')
def ibis8542model_results(ibis8542model_spectra):

    # Load model with random spectra loaded
    m, classifications = ibis8542model_spectra

    # Test with explicit classifications
    result = m.fit(time=range(2), row=range(3), column=range(4), classifications=classifications)

    return result, m, classifications


def assert_results_equal(res1, res2):
    for i in range(len(res1)):
        assert res1[i].parameters == pytest.approx(res2[i].parameters, nan_ok=True)
        assert res1[i].classification == res2[i].classification
        assert res1[i].profile == res2[i].profile
        assert res1[i].success == res2[i].success
        assert np.array_equal(res1[i].index, res2[i].index)


def test_ibis8542model_wrong_length_of_classifications(ibis8542model_spectra):

    # Load model with random spectra loaded
    m, classifications = ibis8542model_spectra

    assert classifications.shape == (2, 3, 4)  # This is what the test needs to work (verifies fixture)

    # Test with too few classifications
    c1 = classifications[:, :, [0, 1]]
    assert c1.shape == (2, 3, 2)
    c2 = classifications[:, [0, 1]]
    assert c2.shape == (2, 2, 4)
    c3 = classifications[[1]]
    assert c3.shape == (1, 3, 4)
    c_wrong_shape = np.transpose(classifications)  # ...but correct size so would not fail otherwise
    assert c_wrong_shape.shape == (4, 3, 2)
    for c in [c1, c2, c3, c_wrong_shape]:
        with pytest.raises(ValueError) as e:
            m.fit(time=range(2), row=range(3), column=range(4), classifications=c)
        assert 'classifications do not match' in str(e.value)

    # Test with too many classifications
    for t, r, c in [
        # (range(2), range(3), range(4)),  # Correct values
        (1, range(3), range(4)),
        (range(2), range(1, 3), range(4)),
        (range(2), range(3), range(2, 4)),
        (range(2), range(3), 3),
        (0, 1, 2),
    ]:
        with pytest.raises(ValueError) as e:
            m.fit(time=t, row=r, column=c, classifications=classifications)
        assert 'classifications do not match' in str(e.value)

    # Test with dimensions of length 1 removed (res1 and res2 should be equivalent)

    c = classifications[:, np.newaxis, 0]
    assert c.shape == (2, 1, 4)
    res1 = m.fit(time=range(2), row=range(1), column=range(4), classifications=c)

    c = classifications[:, 0]
    assert c.shape == (2, 4)
    res2 = m.fit(time=range(2), row=range(1), column=range(4), classifications=c)

    assert len(res1) == len(res2) == 2 * 1 * 4
    assert_results_equal(res1, res2)


@pytest.fixture()
def ibis8542model_resultsobjs(ibis8542model_results):

    def fits2array(results):

        results0 = FitResults((3, 4), 8, time=0)
        results1 = FitResults((3, 4), 8, time=1)

        for i in range(len(results)):

            if results[i].index[0] == 0:
                results0.append(results[i])
            elif results[i].index[0] == 1:
                results1.append(results[i])
            else:
                raise ValueError("invalid time")

        return results0, results1

    return fits2array


def test_ibis8542model_fit(ibis8542model_results, ibis8542model_resultsobjs):

    # # METHOD 1: Test with explicit classifications
    res1, m, classifications = ibis8542model_results

    # # METHOD 2: Test with dummy classifier
    m.neural_network = DummyClassifier(trained=True, classifications=classifications)
    res2 = m.fit(time=range(2), row=range(3), column=range(4))

    assert len(res1) == len(res2) == 2*3*4
    assert_results_equal(res1, res2)

    # # METHOD 3: Test over 4 processing pools
    res3 = m.fit(time=range(2), row=range(3), column=range(4), n_pools=4)
    # (n_pools must be an integer)
    with pytest.raises(TypeError) as e:
        m.fit(time=range(2), row=range(3), column=range(4), n_pools=float(4))
    assert 'n_pools must be an integer' in str(e.value)

    # # Test that FitResults objects can be created consistently for all of the methods

    # Create the FitResults objects from the list of FitResult objects
    results10, results11 = ibis8542model_resultsobjs(res1)  # time : 0, 1
    results20, results21 = ibis8542model_resultsobjs(res2)
    results30, results31 = ibis8542model_resultsobjs(res3)

    # Compare METHOD 1, METHOD 2, METHOD 3
    for t0, t1 in [(results20, results21), (results30, results31)]:  # For each alternative method (tn)
        for results, tn in [(results10, t0), (results11, t1)]:  # Compare to main method (results)
            assert results.parameters == pytest.approx(tn.parameters, nan_ok=True)
            assert np.array_equal(results.profile, tn.profile)
            assert np.array_equal(results.classifications, tn.classifications)
            assert np.array_equal(results.success, tn.success)

    # # Test current values against expected values

    truth10, truth11 = [fits.open(data_path(f)) for f in
                        ("ibis8542model_fit_results10.fits", "ibis8542model_fit_results11.fits")]

    for results, truth in [(results10, truth10), (results11, truth11)]:

        assert results.parameters == pytest.approx(truth['PARAMETERS'].data, nan_ok=True)

        # Compress profile array to integers
        p_uniq = np.unique(results.profile)
        p = np.full_like(results.profile, -1, dtype=int)
        for i in range(len(p_uniq)):
            p[results.profile == p_uniq[i]] = i

        assert np.array_equal(p, truth['PROFILE'].data)
        assert np.array_equal(results.classifications, truth['CLASSIFICATIONS'].data)
        assert np.array_equal(results.success, truth['SUCCESS'].data)


@pytest.mark.parametrize("i,j,k", [(0, 1, 3), (1, 0, 3)])
@figure_test
def test_ibis8542model_plot_indices(pytestconfig, i, j, k, ibis8542model_spectra):
    m = ibis8542model_spectra[0]
    ax = plt.gca()
    m.plot(time=i, row=j, column=k, ax=ax)


@pytest.mark.parametrize("i", np.hstack([range(4, 8), range(18, 20)]))
@figure_test
def test_ibis8542model_plot(pytestconfig, i, ibis8542model_results):
    res, m, _ = ibis8542model_results
    m.plot(fit=res[i])


@pytest.mark.parametrize("i", np.hstack([range(4, 8), range(18, 20)]))
@figure_test
def test_ibis8542model_plot_separate(pytestconfig, i, ibis8542model_results):
    res, m, _ = ibis8542model_results
    m.plot_separate(fit=res[i])


@pytest.mark.parametrize("i", np.hstack([range(4, 8), range(18, 20)]))
@figure_test
def test_ibis8542model_plot_subtraction(pytestconfig, i, ibis8542model_results):
    res, m, _ = ibis8542model_results
    m.plot_subtraction(fit=res[i])


@pytest.mark.parametrize("i", np.hstack([range(4, 8), range(18, 20)]))
@figure_test
def test_ibis8542model_fitresult_plot(pytestconfig, i, ibis8542model_results):
    res, m, _ = ibis8542model_results
    res[i].plot(m)


@figure_test
def test_ibis8542model_plot_spectrum(pytestconfig, ibis8542model_spectra):
    m = ibis8542model_spectra[0]
    spectrum = np.linspace(-1, 1, len(m.constant_wavelengths))
    spectrum = 11 * np.exp(-spectrum ** 2 / 0.05)

    # Raises invalid spectrum dimensions
    with pytest.raises(ValueError) as e:
        m.plot(spectrum=np.array([spectrum, spectrum]))
    assert 'spectrum must have one dimension' in str(e.value)

    # Raises invalid index
    with pytest.raises(IndexError):
        m.plot(time=0, row=10.25, column=0)

    m.plot(spectrum=spectrum, fit=[5, 8542, 0.1, 0.1])


@figure_test
def test_ibis8542model_plot_no_intensity(pytestconfig, ibis8542model_results):
    res, m, _ = ibis8542model_results
    m.plot(fit=res[4], show_intensity=False, show_legend=False)


def test_ibis8542model_save(ibis8542model_results, ibis8542model_resultsobjs, tmp_path):

    # Testing mcalf.models.FitResults.save method

    res1, m = ibis8542model_results[:2]
    results10, results11 = ibis8542model_resultsobjs(res1)

    tmp10 = os.path.join(tmp_path, "ibis8542model_fit_save_results10.fits")
    results10.save(tmp10, m)
    tmp11 = os.path.join(tmp_path, "ibis8542model_fit_save_results11.fits")
    results11.save(tmp11, m)

    diff_kwargs = {
        'ignore_keywords': ['CHECKSUM', 'DATASUM', 'VERSION'],
        'atol': 1e-6,
        'rtol': 1e-5,  # 1e-6 was failing at results10 CHI2[2, 2] on macOS CI env (but nowhere else)
    }
    for saved, truth in [(tmp10, "ibis8542model_fit_results10.fits"),
                         (tmp11, "ibis8542model_fit_results11.fits")]:
        saved = fits.open(saved)
        truth = fits.open(data_path(truth))
        diff = fits.FITSDiff(saved, truth, **diff_kwargs)
        if not diff.identical:  # If this fails tolerances *may* need to be adjusted
            fits.printdiff(saved, truth, **diff_kwargs)
            raise ValueError(f"{saved.filename()} and {truth.filename()} differ")


def test_random_state():

    # Testing that the `random_state` kwarg works as expected on the system

    # Arbitrary wavelength wavelength points
    wavelengths = np.linspace(8541, 8544, 49)

    # Initialise model
    model = IBIS8542Model(original_wavelengths=wavelengths, random_state=0)

    # Get sample classifications
    X, y = make_classification(200, 49, n_classes=5, n_informative=4, random_state=0)

    # Training #1
    model.train(X[::2], y[::2])
    score_a = cross_val_score(model.neural_network, X[1::2], y[1::2])

    # Training #2
    model.train(X[::2], y[::2])
    score_b = cross_val_score(model.neural_network, X[1::2], y[1::2])

    assert score_b == pytest.approx(score_a)
    assert score_b == pytest.approx(np.array([0.45, 0.35, 0.45, 0.45, 0.35]))



import numpy as np
import pytest

from mcalf.models import ModelBase as DummyModel
from mcalf.utils.spec import generate_sigma, normalise_spectrum, reinterpolate_spectrum


def test_reinterpolate_spectrum():
    np.random.seed(0)  # Produce identical results
    x_orig = np.cumsum((np.random.rand(30) * 2) + 1) - 30
    y = np.random.rand(len(x_orig))
    x_const = np.linspace(-20, 20, 15)
    res = reinterpolate_spectrum(y, x_orig, x_const)
    truth = [0.17333763, 0.59316857, 0.55491608, 0.75624398, 0.84520129,
             0.40598020, 0.47206980, 0.40619915, 0.55777817, 0.08496342,
             0.19650506, 0.32022676, 0.47101510, 0.45437529, 0.85481038]
    assert res == pytest.approx(truth)


def test_normalise_spectrum():
    x_orig = np.linspace(-20, 20, 15)
    x_const = x_orig * 0.9
    np.random.seed(0)  # Produce identical results
    y = (np.random.rand(len(x_orig)) * 100) - 50
    prefilter = 1 - (np.random.rand(len(x_orig)) * 0.3)
    res = normalise_spectrum(y, original_wavelengths=x_orig, constant_wavelengths=x_const,
                             prefilter_response=prefilter)
    truth = [0.40058047, 0.29610272, 0.24341921, 0.03779960, 0.07715727,
             0.32041000, 0.00441682, 0.71697103, 1.00000000, 0.00000000,
             0.38235223, 0.35785920, 0.02479681, 0.48465846, 0.68644184]
    assert res == pytest.approx(truth)

    with pytest.raises(ValueError):
        normalise_spectrum(y, original_wavelengths=x_orig)
    with pytest.raises(ValueError):
        normalise_spectrum(y, constant_wavelengths=x_const)


def test_normalise_spectrum_model():
    m = DummyModel(original_wavelengths=[0.0, 0.1])
    m.original_wavelengths = np.linspace(-20, 20, 15)
    m.constant_wavelengths = m.original_wavelengths * 0.9
    np.random.seed(0)  # Produce identical results
    y = (np.random.rand(len(m.original_wavelengths)) * 100) - 50
    m.prefilter_response = 1 - (np.random.rand(len(m.original_wavelengths)) * 0.3)
    res = normalise_spectrum(y, model=m)
    truth = [0.40058047, 0.29610272, 0.24341921, 0.03779960, 0.07715727,
             0.32041000, 0.00441682, 0.71697103, 1.00000000, 0.00000000,
             0.38235223, 0.35785920, 0.02479681, 0.48465846, 0.68644184]
    assert res == pytest.approx(truth)


def test_generate_sigma():
    # Will fail if default parameters of mcalf.utils.spec.generate_sigma change

    res = generate_sigma(1, np.linspace(6.21, 7.24, num=25), 6.743)[::3]
    truth = [0.99999996, 0.96730289, 0.14147828, 0.14147828, 0.14147828,
             0.14147828, 0.14147828, 0.77445118, 0.99999459]
    assert res == pytest.approx(truth)

    res = generate_sigma(2, np.linspace(6.21, 7.24, num=25), 6.743)[::3]
    truth = [0.99999996, 0.96730289, 0.14147828, 0.40000000, 0.40000000,
             0.40000000, 0.14147828, 0.77445118, 0.99999459]
    assert res == pytest.approx(truth)



#!/usr/bin/env python

import os
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


class build_ext(build_ext):

    def get_export_symbols(self, ext):
        if isinstance(ext, CTypes):
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if isinstance(self.ext_map[ext_name], CTypes):
            # Ensure that the extension ends in ".so"
            # Modified version of parent method
            from setuptools.command.build_ext import get_config_var
            ext_suffix = get_config_var('EXT_SUFFIX')
            expanded_suffix = ext_suffix.split('.')
            expanded_suffix[-1] = "so"
            ext_suffix = ".".join(expanded_suffix)
            ext_path = ext_name.split('.')
            return os.path.join(*ext_path) + ext_suffix
        return super().get_ext_filename(ext_name)


class CTypes(Extension):
    pass


is_nt = '_nt' if os.name == 'nt' else ''

setup(
    use_scm_version={'write_to': os.path.join('src', 'mcalf', '_version.py')},
    setup_requires=['setuptools_scm'],
    ext_modules=[CTypes("mcalf.profiles.ext_voigtlib", ["cextern/voigt{}.c".format(is_nt)])],
    cmdclass={'build_ext': build_ext},
)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec

from mcalf.utils.plot import calculate_extent, class_cmap
from mcalf.utils.smooth import mask_classifications

__all__ = ['plot_classifications', 'bar', 'plot_class_map', 'init_class_data']


def plot_classifications(spectra, labels, nrows=None, ncols=None, nlines=20, style='original', cmap=None,
                         show_labels=True, plot_settings={}, fig=None):
    """Plot spectra grouped by their labelled classification.

    Parameters
    ----------
    spectra : ndarray, ndim=2
        Two-dimensional array with dimensions [spectra, wavelengths].
    labels : ndarray, ndim=1, length of `spectra`
        List of classifications for each spectrum in `spectra`.
    nrows : int, optional, default=None
        Number of rows. Defaults to rows of max width 3 axes.
        Special case: four plots will be in a 2x2 grid.
        Only one of `nrows` and `ncols` can be specified.
    ncols : int, optional, default=None
        Number of columns. Defaults to rows of max width 3 axes.
        Special case: four plots will be in a 2x2 grid.
        Only one of `nrows` and `ncols` can be specified.
    nlines : int, optional, default=20
        Maximum number of lines per classification plot.
    style : str, optional, default='original'
        The named matplotlib colormap to extract a :class:`~matplotlib.colors.ListedColormap`
        from. Colours are selected from `vmin` to `vmax` at equidistant values
        in the range [0, 1]. The :class:`~matplotlib.colors.ListedColormap`
        produced will also show bad classifications and classifications
        out of range in grey.
        The default 'original' is a special case used since early versions
        of this code. It is a hardcoded list of 5 colours. When the number
        of classifications exceeds 5, ``style='viridis'`` will be used.
    cmap : callable, optional, default=None
        Function that returns a colour for each input from zero to num. classifications.
        This parameter overrides any cmap requested via the `style` parameter.
        Return value is passed to the `color` parameter of
        :func:`matplotlib.pyplot.axes.Axes.plot`.
    show_labels : bool, optional, default=True
        Whether to label the axes with the corresponding classifications.
    plot_settings : dict, optional, default={}
        Dictionary of keyword arguments to pass to :func:`matplotlib.pyplot.axes.Axes.plot`.
    fig : matplotlib.figure.Figure, optional, default=None
        Figure into which the classifications will be plotted.
        Defaults to the current figure.

    Returns
    -------
    gs : matplotlib.gridspec.GridSpec
        The grid layout subplots are placed on within the figure.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.plot_classifications
    """
    if fig is None:
        fig = plt.gcf()

    # Validate parameters

    for n, v in (('spectra', spectra), ('labels', labels)):
        if not isinstance(v, np.ndarray):
            raise TypeError(f'`{n}` must be a numpy.ndarray, got {type(v)}.')

    if not spectra.ndim == 2:
        raise TypeError('`spectra` must be a 2D array.')

    if not labels.ndim == 1 or not issubclass(labels.dtype.type, np.integer):
        raise TypeError('`labels` must be a 1D array of integers.')

    if len(spectra) != len(labels):
        raise ValueError('`spectra` and `labels` must be the same length along the first dimension.')

    if nrows is not None and ncols is not None:
        raise ValueError('Both `nrows` and `ncols` cannot be given together.')

    for n, v in (('nrows', nrows), ('ncols', ncols)):
        if v is not None and not isinstance(v, (int, np.integer)):
            raise TypeError(f'`{n}` must be an integer, got {type(v)}.')

    if not isinstance(nlines, (int, np.integer)) or nlines <= 0:
        raise TypeError('`nlines` must be a positive integer.')

    # Find and count unique classifications
    classifications = np.unique(labels)
    n = len(classifications)  # number of subplots
    if n == 0:
        return None

    # Set `nrows` and `ncols`
    if ncols is None and nrows is None:
        if n == 1:
            ncols = 1
        elif n == 2 or n == 4:
            ncols = 2
        else:
            ncols = 3
        nrows = int(np.ceil(n / ncols))
    elif ncols is None:
        ncols = int(np.ceil(n / nrows))
    elif nrows is None:
        nrows = int(np.ceil(n / ncols))

    # Verify `nrows` and `ncols`
    if (nrows - 1) * ncols >= n:
        raise ValueError('`nrows` is larger than it needs to be.')
    if nrows * (ncols - 1) >= n:
        raise ValueError('`ncols` is larger than it needs to be.')

    gs = GridSpec(nrows, ncols, figure=fig, wspace=0)

    # Configure the color map
    if cmap is None:
        cmap = class_cmap(style, n)

    for i in range(n):

        ax = fig.add_subplot(gs[i])

        c = classifications[i]
        lines = spectra[labels == c]
        if len(lines) > nlines:  # crop if too big
            lines = lines[:nlines]

        # Whether to crop the y-axis to [0, 1]
        limit_y = False
        if np.nanmin(lines) > -1e-6 and np.nanmax(lines) < 1 + 1e-6:
            limit_y = True

        color = cmap(i)  # extract the single color from the listed colormap
        for line in lines:
            ax.plot(line, color=color, **plot_settings)

        if limit_y:  # if data within range [0, 1]
            ax.set_ylim(0, 1)
            ax.set_yticks([0, 1])  # only show that intensity is scaled [0, 1]

        ax.set_xticks([])  # no wavelengths plotted
        ax.margins(0)  # no "gaps" at line ends

        for loc, spine in ax.spines.items():
            if loc != 'left':
                spine.set_color('none')  # don't draw spine

        if show_labels:
            ax.set_title(f'classification {str(c)}')

    return gs


def bar(class_map=None, vmin=None, vmax=None, reduce=True, style='original', cmap=None, ax=None, data=None):
    """Plot a bar chart of the classification abundances.

    Parameters
    ----------
    class_map : numpy.ndarray[int], ndim=2 or 3
        Array of classifications. If the array is three-dimensional, it is assumed
        that the first dimension is time, and a time average classification will be plotted.
        The time average is the most common positive (valid) classification at each pixel.
    vmin : int, optional, default=None
        Minimum classification integer to plot. Must be greater or equal to zero.
        Defaults to min positive integer in `class_map`.
    vmax : int, optional, default=None
        Maximum classification integer to plot. Must be greater than zero.
        Defaults to max positive integer in `class_map`.
    reduce : bool, optional, default=True
        Whether to perform the time average described in `class_map` info.
    style : str, optional, default='original'
        The named matplotlib colormap to extract a :class:`~matplotlib.colors.ListedColormap`
        from. Colours are selected from `vmin` to `vmax` at equidistant values
        in the range [0, 1]. The :class:`~matplotlib.colors.ListedColormap`
        produced will also show bad classifications and classifications
        out of range in grey.
        The default 'original' is a special case used since early versions
        of this code. It is a hardcoded list of 5 colours. When the number
        of classifications exceeds 5, ``style='viridis'`` will be used.
    cmap : str or matplotlib.colors.Colormap, optional, default=None
        Parameter to pass to matplotlib.axes.Axes.imshow. This parameter
        overrides any cmap requested via the `style` parameter.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which the velocity map will be plotted.
        Defaults to the current axis of the current figure.
    data : dict, optional, default=None
        Dictionary of common classification plotting settings generated by
        :func:`init_class_data`. If present, all other parameters are ignored
        except and `ax`.

    Returns
    -------
    b : matplotlib.container.BarContainer
        The object returned by :func:`matplotlib.axes.Axes.bar` after plotting abundances.

    See Also
    --------
    mcalf.models.ModelBase.classify_spectra : Classify spectra.
    mcalf.utils.smooth.average_classification : Average a 3D array of classifications.

    Notes
    -----
    Visualisation assumes that all integers between `vmin` and `vmax` are valid
    classifications, even if they do not appear in `class_map`.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.bar
    """
    if ax is None:
        ax = plt.gca()

    if data is None:
        if class_map is None:  # `class_map` must always be provided
            raise TypeError("bar() missing 1 required positional argument: 'class_map'")
        data = init_class_data(class_map, vmin=vmin, vmax=vmax, reduce=reduce, style=style, cmap=cmap)

    # Count for each classification
    d = data['class_map'].flatten()
    counts = np.array([len(d[d == i]) for i in data['classes']])
    d = counts / len(d) * 100  # Convert to percentage

    b = ax.bar(data['classes'], d, color=data['cmap'](np.arange(len(data['classes']))))

    ax.set(xlabel='classification', ylabel='abundance (%)',
           xticks=data['classes'], xticklabels=data['classes'])

    return b


def plot_class_map(class_map=None, vmin=None, vmax=None, resolution=None, offset=(0, 0), dimension='distance',
                   style='original', cmap=None, show_colorbar=True, colorbar_settings=None, ax=None, data=None):
    """Plot a map of the classifications.

    Parameters
    ----------
    class_map : numpy.ndarray[int], ndim=2 or 3
        Array of classifications. If the array is three-dimensional, it is assumed
        that the first dimension is time, and a time average classification will be plotted.
        The time average is the most common positive (valid) classification at each pixel.
    vmin : int, optional, default=None
        Minimum classification integer to plot. Must be greater or equal to zero.
        Defaults to min positive integer in `class_map`.
    vmax : int, optional, default=None
        Maximum classification integer to plot. Must be greater than zero.
        Defaults to max positive integer in `class_map`.
    resolution : tuple[float] or astropy.units.quantity.Quantity, optional, default=None
        A 2-tuple (x, y) containing the length of each pixel in the x and y direction respectively.
        If a value has type :class:`astropy.units.quantity.Quantity`, its axis label will
        include its attached unit, otherwise the unit will default to Mm.
        If `resolution` is None, both axes will be ticked with the default pixel value
        with no axis labels.
    offset : tuple[float] or int, length=2, optional, default=(0, 0)
        Two offset values (x, y) for the x and y axis respectively.
        Number of pixels from the 0 pixel to the first pixel. Defaults to the first
        pixel being at 0 length units. For example, in a 1000 pixel wide dataset,
        setting offset to -500 would place the 0 Mm location at the centre.
    dimension : str or tuple[str] or list[str], length=2, optional, default='distance'
        If an `ax` (and `resolution`) is provided, use this string as the `dimension name`
        that appears before the ``(unit)`` in the axis label.
        A 2-tuple (x, y) or list [x, y] can instead be given to provide a different name
        for the x-axis and y-axis respectively.
    style : str, optional, default='original'
        The named matplotlib colormap to extract a :class:`~matplotlib.colors.ListedColormap`
        from. Colours are selected from `vmin` to `vmax` at equidistant values
        in the range [0, 1]. The :class:`~matplotlib.colors.ListedColormap`
        produced will also show bad classifications and classifications
        out of range in grey.
        The default 'original' is a special case used since early versions
        of this code. It is a hardcoded list of 5 colours. When the number
        of classifications exceeds 5, ``style='viridis'`` will be used.
    cmap : str or matplotlib.colors.Colormap, optional, default=None
        Parameter to pass to matplotlib.axes.Axes.imshow. This parameter
        overrides any cmap requested via the `style` parameter.
    show_colorbar : bool, optional, default=True
        Whether to draw a colorbar.
    colorbar_settings : dict, optional, default=None
        Dictionary of keyword arguments to pass to :func:`matplotlib.figure.Figure.colorbar`.
        Ignored if `show_colorbar` is False.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which the velocity map will be plotted.
        Defaults to the current axis of the current figure.
    data : dict, optional, default=None
        Dictionary of common classification plotting settings generated by
        :func:`init_class_data`. If present, all other parameters are ignored
        except `show_colorbar` and `ax`.

    Returns
    -------
    im : matplotlib.image.AxesImage
        The object returned by :func:`matplotlib.axes.Axes.imshow` after plotting `class_map`.

    See Also
    --------
    mcalf.models.ModelBase.classify_spectra : Classify spectra.
    mcalf.utils.smooth.average_classification : Average a 3D array of classifications.

    Notes
    -----
    Visualisation assumes that all integers between `vmin` and `vmax` are valid
    classifications, even if they do not appear in `class_map`.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.plot_class_map
    """
    if ax is None:
        ax = plt.gca()

    if data is None:
        if class_map is None:  # `class_map` must always be provided
            raise TypeError("plot_class_map() missing 1 required positional argument: 'class_map'")
        data = init_class_data(class_map, vmin=vmin, vmax=vmax,
                               resolution=resolution, offset=offset, dimension=dimension,
                               style=style, cmap=cmap, colorbar_settings=colorbar_settings, ax=ax)

    # Plot `class_map` (using special imshow `data` keyword)
    im = ax.imshow('class_map', cmap='cmap', vmin='plot_vmin', vmax='plot_vmax',
                   origin='lower', extent='extent', interpolation='nearest',
                   data=data)

    if show_colorbar:
        ax.get_figure().colorbar(im, **data['colorbar_settings'])

    return im


def init_class_data(class_map, vmin=None, vmax=None, reduce=True, resolution=None, offset=(0, 0),
                    dimension='distance', style='original', cmap=None, colorbar_settings=None, ax=None):
    """Initialise dictionary of common classification plotting data.

    Parameters
    ----------
    class_map : numpy.ndarray[int], ndim=2 or 3
        Array of classifications. If `reduce` is True (default) and the array is
        three-dimensional, it is assumed that the first dimension is time, and
        a time average classification will be calculated. The time average is
        the most common positive (valid) classification at each pixel.
    vmin : int, optional, default=None
        Minimum classification integer to include. Must be greater or equal to zero.
        Defaults to min positive integer in `class_map`. Classifications below this
        value will be set to -1.
    vmax : int, optional, default=None
        Maximum classification integer to include. Must be greater than zero.
        Defaults to max positive integer in `class_map`. Classifications above this
        value will be set to -1.
    reduce : bool, optional, default=True
        Whether to perform the time average described in `class_map` info.
    resolution : tuple[float] or astropy.units.quantity.Quantity, optional, default=None
        A 2-tuple (x, y) containing the length of each pixel in the x and y direction respectively.
        If a value has type :class:`astropy.units.quantity.Quantity`, its axis label will
        include its attached unit, otherwise the unit will default to Mm.
        If `resolution` is None, both axes will be ticked with the default pixel value
        with no axis labels.
    offset : tuple[float] or int, length=2, optional, default=(0, 0)
        Two offset values (x, y) for the x and y axis respectively.
        Number of pixels from the 0 pixel to the first pixel. Defaults to the first
        pixel being at 0 length units. For example, in a 1000 pixel wide dataset,
        setting offset to -500 would place the 0 Mm location at the centre.
    dimension : str or tuple[str] or list[str], length=2, optional, default='distance'
        If an `ax` (and `resolution`) is provided, use this string as the `dimension name`
        that appears before the ``(unit)`` in the axis label.
        A 2-tuple (x, y) or list [x, y] can instead be given to provide a different name
        for the x-axis and y-axis respectively.
    style : str, optional, default='original'
        The named matplotlib colormap to extract a :class:`~matplotlib.colors.ListedColormap`
        from. Colours are selected from `vmin` to `vmax` at equidistant values
        in the range [0, 1]. The :class:`~matplotlib.colors.ListedColormap`
        produced will also show bad classifications and classifications
        out of range in grey.
        The default 'original' is a special case used since early versions
        of this code. It is a hardcoded list of 5 colours. When the number
        of classifications exceeds 5, ``style='viridis'`` will be used.
    cmap : str or matplotlib.colors.Colormap, optional, default=None
        Parameter to pass to matplotlib.axes.Axes.imshow. This parameter
        overrides any cmap requested via the `style` parameter.
    colorbar_settings : dict, optional, default=None
        Dictionary of keyword arguments to pass to :func:`matplotlib.figure.Figure.colorbar`.
    ax : matplotlib.axes.Axes, optional, default=None
        Axes into which the classification map will be plotted.
        Defaults to the current axis of the current figure.

    Returns
    -------
    data : dict
        Common classification plotting settings.

    See Also
    --------
    mcalf.visualisation.bar : Plot a bar chart of the classification abundances.
    mcalf.visualisation.plot_class_map : Plot a map of the classifications.
    mcalf.utils.smooth.mask_classifications : Mask 2D and 3D arrays of classifications.
    mcalf.utils.plot.calculate_extent : Calculate the extent from a particular data shape and resolution.
    mcalf.utils.plot.class_cmap : Create a listed colormap for a specific number of classifications.

    Examples
    --------
    .. minigallery:: mcalf.visualisation.init_class_data
    """
    # Mask and average classification map according to the classification range and shape
    class_map, vmin, vmax = mask_classifications(class_map, vmin, vmax, reduce=reduce)

    # Create a list of the classifications
    classes = np.arange(vmin, vmax + 1, dtype=int)

    # Configure the color map
    if cmap is None:
        cmap = class_cmap(style, len(classes))

    # Calculate a specific extent if a resolution is specified
    extent = calculate_extent(class_map.shape, resolution, offset,
                              ax=ax, dimension=dimension)

    # Create and update colorbar settings
    cbar_settings = {'ax': [ax], 'ticks': classes}
    if colorbar_settings is not None:
        cbar_settings.update(colorbar_settings)

    data = {
        'class_map': class_map,
        'vmin': vmin,
        'vmax': vmax,
        'plot_vmin': vmin - 0.5,
        'plot_vmax': vmax + 0.5,
        'classes': classes,
        'extent': extent,
        'cmap': cmap,
        'colorbar_settings': cbar_settings,
    }

    return data



import numpy as np
import scipy.ndimage

__all__ = ['moving_average', 'gaussian_kern_3d', 'smooth_cube', 'mask_classifications']


def moving_average(array, width):
    """Boxcar moving average.

    Calculate the moving average of an array with a boxcar of defined width. An odd width is recommended.

    Parameters
    ----------
    array : numpy.ndarray, ndim=1
        Array to find the moving average of.
    width : int
        Width of the boxcar. Odd integer recommended. Less than or equal to length of `array`.

    Returns
    -------
    averaged : numpy.ndarray, shape=`array`
        Averaged array.

    Notes
    -----
    The moving average is calculated at each point of the `array` by finding the (unweighted) mean of the subarrays
    of length given by `width`. These subarrays are centred at the point in the `array` that the current average is
    currently being calculated for. If an odd `width` is chosen, the sub array will include the current point plus an
    equal number of points on either side. However, if an even `width` is chosen, the sub array will bias including the
    extra point to the left of the current index. If the subarray spans past the boundaries, the values beyond the
    boundary is ignored and the mean is calculated by dividing by the number of points that are inside the boundaries.

    Examples
    --------
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> moving_average(x, 3)
    array([1.5, 2. , 3. , 4. , 4.5])
    >>> moving_average(x, 2)
    array([1. , 1.5, 2.5, 3.5, 4.5])
    """

    if not isinstance(width, (int, np.integer)) or width <= 0 or width > len(array):
        raise ValueError("`width` must be a positive integer less than the length of `array`, got %s." % width)

    kernel = np.ones(width) / width
    averaged = np.convolve(array, kernel, mode='same')

    # Apply corrections at edge
    to_correct_left = int(width/2)
    to_correct_right = to_correct_left
    to_correct_right -= 0 if width % 2 else 1

    # Adjust weighting on left edge
    n_overlap = width - 1
    for i in np.arange(to_correct_left-1, -1, -1):
        averaged[i] *= width / n_overlap
        n_overlap -= 1

    # Adjust weighting on right edge
    n_overlap = width - 1
    for i in range(len(array) - to_correct_right, len(array)):
        averaged[i] *= width / n_overlap
        n_overlap -= 1

    return averaged


def gaussian_kern_3d(width=5, sigma=(1, 1, 1)):
    """3D Gaussian kernel.

    Create a Gaussian kernel of shape `width`*`width`*`width`.

    Parameters
    ----------
    width : int, optional, default=5
        Length of all three dimensions of the Gaussian kernel.
    sigma : array_like, tuple, optional, default=(1, 1, 1)
        Sigma values for the time, horizontal and vertical dimensions.

    Returns
    -------
    kernel : numpy.ndarray, shape=(`width`, `width`, `width`)
        The generated kernel.

    Examples
    --------
    >>> gaussian_kern_3d(width=3, sigma=(2, 1, 1.5))
    array([[[0.42860385, 0.53526143, 0.42860385],
            [0.48567179, 0.60653066, 0.48567179],
            [0.42860385, 0.53526143, 0.42860385]],
    <BLANKLINE>
           [[0.70664828, 0.8824969 , 0.70664828],
            [0.8007374 , 1.        , 0.8007374 ],
            [0.70664828, 0.8824969 , 0.70664828]],
    <BLANKLINE>
           [[0.42860385, 0.53526143, 0.42860385],
            [0.48567179, 0.60653066, 0.48567179],
            [0.42860385, 0.53526143, 0.42860385]]])
    """
    s = np.linspace(-1, 1, width)
    x, y, z = np.meshgrid(s, s, s)
    kernel = np.exp(-x**2/(2*sigma[0]**2) - y**2/(2*sigma[1]**2) - z**2/(2*sigma[2]**2))
    return kernel


def smooth_cube(cube, mask, **kwargs):
    """Apply Gaussian smoothing to velocities.

    Smooth the cube of velocities with a Gaussian kernel, applying weights at boundaries.

    Parameters
    ----------
    cube : numpy.ndarray, ndim=3
        Cube of velocities with dimensions [time, row, column].
    mask : numpy.ndarray, ndim=2
        The mask to apply to the [row, column] at every time. Points that are 0 or false will be removed.
    **kwargs : dict, optional
        Keyword arguments to pass to :func:`gaussian_kern_3d`.

    Returns
    -------
    cube_ : numpy.ndarray, shape=`cube`
        The smoothed cube.
    """
    # Masking
    cube_ = cube.copy()
    mask3d = np.empty_like(cube_)  # Mask for `cube_`
    for timestep in range(len(mask3d)):  # Copy mask throughout time
        mask3d[timestep] = mask.copy()
    cube_[np.isnan(cube_)] = 0  # Remove NaN
    cube_[mask3d == 0] = 0  # Remove masked points

    kernel = gaussian_kern_3d(**kwargs)  # Get the kernel

    # Count the number of neighbours (weighting) at each pixel
    neighbour_count = scipy.ndimage.convolve(mask3d, kernel, mode="constant")
    m = neighbour_count > 0  # Record the points that do have neighbours

    cube_ = scipy.ndimage.convolve(cube_, kernel, mode="constant")  # Apply smoothing
    cube_[m] /= neighbour_count[m]  # Normalise depending on neighbours

    cube_[mask3d == 0] = np.nan  # Restore masked pixels
    cube_[np.isnan(cube)] = np.nan  # Restore NaN values

    return cube_


def mask_classifications(class_map, vmin=None, vmax=None, reduce=True):
    """Mask 2D and 3D arrays of classifications.

    If 3D, also reduces to 2D by selecting the most common classification
    along the first dimension.

    Parameters
    ----------
    class_map : numpy.ndarray[int], ndim=2 or 3
        Array of classifications. If `reduce` is True (default) and the array is
        three-dimensional, it is assumed that the first dimension is time, and
        a time average classification will be calculated. The time average is
        the most common positive (valid) classification at each pixel.
    vmin : int, optional, default=None
        Minimum classification integer to include. Must be greater or equal to zero.
        Defaults to min positive integer in `class_map`. Classifications below this
        value will be set to -1.
    vmax : int, optional, default=None
        Maximum classification integer to include. Must be greater than zero.
        Defaults to max positive integer in `class_map`. Classifications above this
        value will be set to -1.
    reduce : bool, optional, default=True
        Whether to perform the time average described in `class_map` info.

    Returns
    -------
    class_map : numpy.ndarray[int], ndim=2
        `class_map` with values between `vmin` and `vmax`
        averaged along the first dimension.
    vmin : int
        Updated `vmin` value.
    vmax : int
        Updated `vmax` value.

    See Also
    --------
    mcalf.visualisation.plot_class_map : Plot a map of the classifications.
    """
    # Validate the `class_map` parameter
    if not isinstance(class_map, np.ndarray):
        raise TypeError(f'`class_map` must be a numpy.ndarray, got {type(class_map)}.')
    if class_map.ndim not in (2, 3):
        raise ValueError(f'`class_map` must have either 2 or 3 dimensions, got {class_map.ndim}.')
    if not issubclass(class_map.dtype.type, np.integer):
        raise TypeError(f'`class_map` must be an array of integers, got {class_map.dtype}.')

    # Short-circuit if all classifications negative
    if len(class_map[class_map >= 0]) == 0:
        class_map = class_map.copy()
        if reduce and class_map.ndim == 3:
            class_map = class_map[0]
        class_map[:] = -1  # Set all invalid to -1 for consistency with main code
        if vmin is None:
            vmin = 0
        if vmax is None:
            vmax = 0
        return class_map, vmin, vmax

    # Validate or set `vmin` and `vmax`
    if vmax is None:
        vmax = np.max(class_map[class_map >= 0])
    elif not isinstance(vmax, (int, np.integer)):
        raise TypeError(f'`vmax` must be an integer, got {type(vmax)}.')
    elif vmax < 0:
        raise ValueError('`vmax` must not be less than zero.')
    if vmin is None:
        vmin = np.min(class_map[class_map >= 0])
    elif not isinstance(vmin, (int, np.integer)):
        raise TypeError(f'`vmin` must be an integer, got {type(vmin)}.')
    elif vmin < 0:
        raise ValueError('`vmin` must not be less than zero.')
    if vmin > vmax:
        raise ValueError(f'`vmin` must be less than `vmax`, got {vmin} to {vmax}.')

    # Ignore classifications outside of range
    class_map = class_map.copy()
    class_map[class_map < vmin] = -1
    class_map[class_map > vmax] = -1

    # If 3D, choose the most common classification along the first dimension
    if reduce and class_map.ndim == 3:

        c = np.arange(vmin, vmax + 1)  # classes
        bins = np.arange(vmin - 0.5, vmax + 1)  # len(bins)=len(c)+1

        ave = np.empty(class_map.shape[1:], dtype=int)  # init

        for i in range(len(class_map[0])):  # spatial rows
            for j in range(len(class_map[0, 0])):  # spatial columns
                counts = np.histogram(class_map[:, i, j], bins)[0]
                if counts.max() == 0:
                    ave[i, j] = -1  # no valid classifications
                else:
                    ave[i, j] = c[counts.argmax()]

        class_map = ave  # Replace 3D array with 2D average

    return class_map, vmin, vmax
